/*

COPYRIGHT © 2009 THE OHIO STATE UNIVERSITY
ALL RIGHTS RESERVED

PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR NONCOMMERCIAL
EDUCATION AND RESEARCH PURPOSES, SO LONG AS NO FEE IS CHARGED, AND SO
LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION, AND THE
DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS THE NAME OF
THE OHIO STATE UNIVERSITY IS NOT USED IN ANY ADVERTISING OR PUBLICITY
PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE WITHOUT
SPECIFIC, WRITTEN PRIOR AUTHORIZATION.

THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE OHIO
STATE UNIVERSITY AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
WARRANTY BY THE OHIO STATE UNIVERSITY OF ANY KIND, EITHER EXPRESS OR
IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE OHIO STATE
UNIVERSITY SHALL NOT BE LIABLE FOR ANY DAMAGES, INCLUDING SPECIAL,
INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY
CLAIM ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE,
EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
DAMAGES.
*/

/* From Robert Forties */

//Calculate the observables using the FRET distribution P(E|m) and fixed protein binding parameters 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include "fold.h"
#include "part_func.h"
#include "fold_vars.h"
#include "PS_dot.h"
#include "utils.h"
#include "energy_const.h"

#define VERBOSE 1
void m_to_length(FLT_OR_DBL * P_x, FLT_OR_DBL * P_m, float xmin, float xmax, int nbins, int n);
void length_to_FRET(FLT_OR_DBL * P_x, FLT_OR_DBL * P_E, int nbins, float xmin, float dx, int nbinsx);
float m_to_FRET(FLT_OR_DBL * P_E, FLT_OR_DBL * P_m, FLT_OR_DBL * E_m, float xmin, float xmax, int nbins, int n);
float m_to_FRET_dist(FLT_OR_DBL * P_E, FLT_OR_DBL * P_m, FLT_OR_DBL * E_m, float xmin, float xmax, int nbins, int n);
void print_NC_energies ();
double Calc_Avg_N ( char * Seq, char * struct1, double kT );
double Calc_N_i ( char * Seq, char * struct1 , double kT, double * Any_NC_SS);

extern void  read_parameter_file(const char fname[]);

/*@unused@*/
static char rcsid[] = "$Id: RNAfold.c,v 1.3 2003/12/17 13:22:51 rbund Exp $";

#define PRIVATE static

static char  scale1[] = "....,....1....,....2....,....3....,....4";
static char  scale2[] = "....,....5....,....6....,....7....,....8";

int display_fx = 0;
int display_PE = 1;
int calc_PE_C = 0;
int calc_E_C = 0; //calculate free energy as a function of [NC]
int calc_N_i = 0; //calculate sequence dependent NC binding energy
int calc_N_i_C = 0; //calculate sequence dependent NC binding energy
int print_N_i_C_dist =0; //print distribution of N_i_C
char * N_i_C_FileName = "N_i_C_dist_test.dat";
int dm = 0;


PRIVATE void usage(void);

/*--------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  char *string, *line;
  char *structure=NULL, *cstruc=NULL;
  char  fname[13], ffname[2000], gfname[2000];
  char  *ParamFile=NULL;
  char  *ns_bases=NULL, *c;
  int   length, sym, r;
  int fnameindex;
  double energy, min_en;
  double sfact=1.07;
  int   pf=0, istty;
  int noconv=0;
   
  string=NULL;
  fnameindex=0;

  //variables for calculation of distributions
  force = 0.0;
  do_backtrack=0;
  do_autoscale=0;
  do_fixedextension=1;
  do_polymer=0;
  salt = 0.097;// 0.05;0.0004;
  temperature = 21;
  double kT = (temperature+K0)*GASCONST/1000.0;   /* kT in cal/mol  */

  footprintDS = 6;
  double E_NC_DS0 = +100.0;//	-9.52426;//-11.509519;
  exp_E_NC_DS = exp (-E_NC_DS0 / footprintDS / kT);

  R_FRET = 6.0; 
  High_NC_energy = 0.0;//-.266;//-0.251807;//-1.194123;//-9.967671;//-9.957689;//10.0;//-12.13 ; //-7.5;10.0;
  High_NC_pos = 11;
  int MaxLength =100; /*maximum length for sequences to be folded*/

  double dmu_NC_end = 0.0;
  int end_length = 6;

  double dmu = 0.01;
  double dC = 10E-9; //concentration step scale
  int N_C = 100; //number of different NC concentrations to calculate

  //double C_N_i[] = {20,40,50,60,80,100,120,160,200,300,400,500,600,800,1600};
  double C_N_i[] = {0,1,2,4,6,8,10,12,16,20,30,40,50,60,80,100,120,140,160,180,200,250,300,400,500,600,700,800,900,1000,1200,1500,2000,3000,4000,5000};
  int N_i = 3;

  //parameters for FRET distribution output
  int nbins = 4000;
  float xmin = 0.;
  float xmax = 20;
  int nbinsF = 50;

  int i,j,k,l, nless; /*counters*/

  //output file
  char * OutFileName = "output.dat";
  if (argc > 1) OutFileName = argv[1];
  FILE * OutFile = fopen(OutFileName, "w");

  FILE * N_i_C_File;
  if (print_N_i_C_dist)  //initialize N_i_C_File
  {
	N_i_C_File = fopen(N_i_C_FileName, "w");
        fprintf(N_i_C_File, "#  C_NC      N    P(C_NC,N)\n");
	fclose(N_i_C_File);
  }

  printf("High_NC_energy=%f\n",High_NC_energy);
  printf("E_NC_DS=%f\n",E_NC_DS0);
  printf("footprintDS=%f\n",footprintDS);

  double * pr_i; /*probability base i is paired*/
  pr_i = malloc((MaxLength+1)*sizeof(double)); 

  double E_FRET; //Calculated average FRET efficiency 

  char *SeqEnd;  /*pointer to end of DNA sequence*/
  float Freq;/*gene freq*/

  char * struct1; /*structure containing pairing data for current seq*/
  int n; /*length of current sequence*/
  int nseq; /*number of sequence currently used*/
  char * Seq; /*current sequence*/

  float extension; /*initialize DNA extension*/
  float segment_length = 1.0; /*entension per basepair*/


  FILE * Heatmap_File; //Output file for heatmap
  char Heatmap_File_Name [100];  //Name of Output file for heatmap

  //deviation from average binding energy for NC bound starting at each base
  //exp_dmu = malloc (MaxLength*sizeof(double));
  //for (i=0;i<MaxLength;i++) exp_dmu[i] = 25.; //initialize to 1

  float Energy;/*energy of sequence*/


  FLT_OR_DBL * P_x = (FLT_OR_DBL* ) space(sizeof(FLT_OR_DBL)*(nbins+1));
  FLT_OR_DBL * P_m = (FLT_OR_DBL* ) space(sizeof(FLT_OR_DBL)*(MaxLength+1));
  FLT_OR_DBL * P_E = (FLT_OR_DBL* ) space(sizeof(FLT_OR_DBL)*(nbinsF+1));
  FLT_OR_DBL * E_m = (FLT_OR_DBL* ) space(sizeof(FLT_OR_DBL)*(MaxLength+1));

  double avg_melting; //average melting probability


  double Avg_N;
  double Any_NC_bound;
  double Any_NC_SS;
  double Energy0;  

  //end variables for calculation of distributions

  for (i=1; i<argc; i++) {
    if (argv[i][0]=='-') 
      switch ( argv[i][1] )
	{
	case 'T':  if (argv[i][2]!='\0') usage();
	  if(i==argc-1) usage();
	  r=sscanf(argv[++i], "%lf", &temperature);
	  if (!r) usage();
	  break;
	case 'F':  if (argv[i][2]!='\0') usage();
	  if(i==argc-1) usage();
	  r=sscanf(argv[++i], "%f", &force);
	  if (!r) usage();
	  break;
	case 'f':  if (argv[i][2]!='\0') usage();
	  if(i==argc-1) usage();
	  fnameindex=++i;
	  break;
	case 'p':  pf=1;
	  if (argv[i][2]!='\0')
	    (void) sscanf(argv[i]+2, "%d", &do_backtrack);
	  break;
	case 'n':
	  if ( strcmp(argv[i], "-noGU")==0) noGU=1;
	  if ( strcmp(argv[i], "-noCloseGU")==0) no_closingGU=1;
	  if ( strcmp(argv[i], "-noLP")==0) noLonelyPairs=1;
	  if ( strcmp(argv[i], "-nsp") ==0) {
            if (i==argc-1) usage();
	    ns_bases = argv[++i];
	  }
	  if ( strcmp(argv[i], "-noconv")==0) noconv=1;
	  break;
	case '4':
	  tetra_loop=0;
	  break;
	case 'e':
	  if(i==argc-1) usage();
	  r=sscanf(argv[++i],"%d", &energy_set);
	  if (!r) usage();
	  break;
	case 'C':
	  fold_constrained=1;
	  break;
	case 'S':
	  if(i==argc-1) usage();
	  r=sscanf(argv[++i],"%lf", &sfact);
          do_autoscale=0;
	  if (!r) usage();
	  break;
	case 'd': dangles=0;
	  if (argv[i][2]!='\0') {
	    r=sscanf(argv[i]+2, "%d", &dangles);
	    if (r!=1) usage();
	  }
	  break;
	case 'P':
	  if (i==argc-1) usage();
	  ParamFile = argv[++i];
	  break;
	case 'o':
	  if (i==argc-1) usage();
	  r=sscanf(argv[++i], "%d", &open_bases);
	  if (!r) usage();
	  break;
	case 'v':
	  subtract_one=1;
	  break;
	case 'b':
	  do_polymer=1;
	  break;
	default: usage();
	} 
  }

  if (ParamFile != NULL)
    read_parameter_file(ParamFile);
   
  if (ns_bases != NULL) {
    nonstandards = space(33);
    c=ns_bases;
    i=sym=0;
    if (*c=='-') {
      sym=1; c++;
    }
    while (*c!='\0') {
      if (*c!=',') {
	nonstandards[i++]=*c++;
	nonstandards[i++]=*c;
	if ((sym)&&(*c!=*(c-1))) {
	  nonstandards[i++]=*c;
	  nonstandards[i++]=*(c-1);
	}
      }
      c++;
    }
  }
  istty = isatty(fileno(stdout))&&isatty(fileno(stdin));
  if ((fold_constrained)&&(istty)) {
    printf("Input constraints using the following notation:\n");
    printf("| : paired with another base\n");
    printf(". : no constraint at all\n");
    printf("x : base must not pair\n");
    printf("< : base i is paired with a base j<i\n");
    printf("> : base i is paired with a base j>i\n");
    printf("matching brackets ( ): base i pairs base j\n");
  } 
	
  do {				/* main loop: continue until end of file */
    if (istty) {
      printf("\nInput string (upper or lower case); @ to quit\n");
      printf("%s%s\n", scale1, scale2);
    }
    fname[0]='\0';
    if ((line = get_line(stdin))==NULL) break;

    /* skip comment lines and get filenames */
    while ((*line=='*')||(*line=='\0')||(*line=='>')) {
      if (*line=='>')
	(void) sscanf(line, ">%12s", fname);
      printf("%s\n", line);
      free(line);
      if ((line = get_line(stdin))==NULL) break;
    } 

    if ((line ==NULL) || (strcmp(line, "@") == 0)) break;

    string = (char *) space(strlen(line)+1);
    (void) sscanf(line,"%s",string);
    free(line);
    length = (int) strlen(string);

    structure = (char *) space((unsigned) length+1);
    if (fold_constrained) {
      cstruc = get_line(stdin);
      if (cstruc!=NULL) 
	strncpy(structure, cstruc, length);
      else
	fprintf(stderr, "constraints missing\n");
    }
    for (l = 0; l < length; l++) {
      string[l] = toupper(string[l]);
      if (!noconv && string[l] == 'T') string[l] = 'U';
    }
    if (istty)
      printf("length = %d\n", length);


    init_pf_fold(length);
    print_NC_energies ();

   
    //Code to calculate and display observables   
      Seq = string;
      n = (int) strlen(Seq);
      struct1 = (char* ) space(sizeof(char)*(strlen(Seq)+1)); 
      //exp_dmu = exp_dmuSeq[nseq];


      //printf("%s:  ",SeqName[nseq]);
      //if (VERBOSE) printf("\n %s\n",Seq);

      //printf("pf_fold\n");
      Energy=pf_fold(Seq ,struct1);

      if (do_backtrack)
      { 
	printf("%s\n",struct1);    
	
        for(i=0;i<=n;i++) 
	{
	   pr_i[i]=0.;
	   for(j=0;j<=i;j++)
	   {
	      pr_i[i] += pr[iindx[j]-i];
	      if (!(i==j)) pr_i[j] += pr[iindx[j]-i];
	   }
	}

        //print melting probabilities
	strcpy(Heatmap_File_Name,"");
	strcat(Heatmap_File_Name,"Heatmap");
	//strcat(Heatmap_File_Name,SeqName[nseq]);
	strcat(Heatmap_File_Name,".txt");
	printf("Heatmap_File_Name = %s\n",Heatmap_File_Name);
	Heatmap_File = fopen(Heatmap_File_Name,"w");
        avg_melting=1.0;
	for(i=1;i<=n;i++)
	{ 
	  fprintf(Heatmap_File,"%f\n",1.0-pr_i[i]); 
	  printf("%f\n",1.0-pr_i[i]);
	  avg_melting -= pr_i[i]/n;
	}
	fclose(Heatmap_File);
        printf ("avg melting prob = %f\n", avg_melting);

	for(i=1;i<=n;i++)
	{  
	  //printf("%f\n",pNCr[i]); 
	}

	/*for(i=0;i<=n;i++) {
	   for(j=0;j<=n;j++) {
	        if (i>j) 
	        {
		  k=i;
	          l=j;
                } else {
		  k=j;
	          l=i;
		}
		printf("%f ",pr[iindx[l]-k]);	      
	   }
	   printf("\n");	 
	}*/
      }
      if (VERBOSE) printf("Energy=%f\n",Energy);
      if (do_fixedextension)
      {

	for(i=0;i<=n;i++) P_m[i]=exp((-fx[i]+Energy)/kT);

	m_to_length(P_x, P_m, xmin, xmax, nbins, n) ;    
  
	extension=0.0;
  	for(i=0;i<=n;i++) {
    	  //printf("fx[%d]=%f, P=%f\n",i,fx[i],P_m[i]);
	  //printf("%f\n",exp((-fx[i]+Energy)/kT));
	  extension += (i*segment_length)*exp((-fx[i]+Energy)/kT);
  	}
	if (VERBOSE) printf("extension = %f\n",extension);
    
    	/*for(i=0;i<=nbins;i++) {
    	  printf("P_x[%f]=%f\n",i*(xmax-xmin)/nbins+xmin,P_x[i]);
	}*/

	/*length_to_FRET(P_x, P_E, nbinsF, xmin, (xmax-xmin)/nbins, nbins);

    	for(i=0;i<=nbinsF;i++) {
    	  printf("P_E[%f]=%f\n",i*1.0/nbinsF,P_E[i]);
  	}*/
	if (VERBOSE) printf ("Average E = ");


	E_FRET=m_to_FRET_dist(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
	//printf ("%f, diffsqr=%f\n",E_FRET, (E_FRET-E_Exp[nseq])*(E_FRET-E_Exp[nseq]));

	if (VERBOSE)
	{
	 if (display_fx) for(i=0;i<=n;i++) printf("fx[%d]=%f, P=%f\n",i,fx[i],P_m[i]);
	 if (display_PE)
	 {
          printf("P_E = (%f step size)\n",1.0/nbinsF);
    	  for(i=0;i<nbinsF;i++) {
    	   //printf("P_E[ %f ] = %f\n",i*1.0/nbinsF,P_E[i]);
 	   printf("%f\n",P_E[i]);
  	  }
	 }
	 for(i=0;i<=n;i++) {
    	 // printf("E_m[ %d ] = %f %f\n",i,E_m[i], P_m[i]);
	 }

	}
      }

      if (calc_E_C) //calculate E as a function of NC concentration
      {
	printf("C_NC FRET    Energy    Avg_N\n");
        C_NC = 0E-9;
	Energy=pf_fold(Seq ,struct1);
        Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	E_FRET=m_to_FRET(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
        printf("%g  %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
        C_NC = 1E-9;
	Energy=pf_fold(Seq ,struct1);
        Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	E_FRET=m_to_FRET(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
        printf("%g  %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
        C_NC = 0E-9;
        for (i=0;i<N_C;i++)
        {
          C_NC += dC;
          Energy=pf_fold(Seq ,struct1);
	  for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	  E_FRET=m_to_FRET(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
          Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	  printf("%g %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
        }
      }

      if (calc_PE_C) //calculate P(E) as a function of NC concentration
      {
        fprintf(OutFile, "#P_E = (%f step size)\n",1.0/nbinsF);
	//printf("C_NC FRET    Energy    Avg_N\n");
        C_NC = 0E-9;
	Energy=pf_fold(Seq ,struct1);
        Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	E_FRET=m_to_FRET_dist(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
        //printf("%g  %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
    	for(i=0;i<nbinsF+1;i++) {
    	   //printf("P_E[ %f ] = %f\n",i*1.0/nbinsF,P_E[i]);
 	   fprintf(OutFile, "%f %f %f\n",C_NC*1E9, i*1.0/nbinsF, P_E[i]);
  	  }
          fprintf(OutFile,"\n");
        C_NC = 1E-9;
	Energy=pf_fold(Seq ,struct1);
        Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	E_FRET=m_to_FRET_dist(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
        //printf("%g  %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
    	for(i=0;i<nbinsF+1;i++) {
    	   //printf("P_E[ %f ] = %f\n",i*1.0/nbinsF,P_E[i]);
 	   fprintf(OutFile, "%f %f %f\n",C_NC*1E9, i*1.0/nbinsF, P_E[i]);
  	  }
          fprintf(OutFile,"\n");
        C_NC = 0E-9;
        for (i=0;i<N_C;i++)
        {
          C_NC += dC;
          Energy=pf_fold(Seq ,struct1);
	  for(j=0;j<=n;j++) P_m[j]=exp((-fx[j]+Energy)/kT);
	  E_FRET=m_to_FRET_dist(P_E, P_m,E_m, xmin, xmax, nbinsF, n);
          Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	  //printf("%g %f %f %f\n",C_NC*1E9,E_FRET, Energy,Avg_N );
    	  for(j=0;j<nbinsF+1;j++) {
    	   //printf("P_E[ %f ] = %f\n",i*1.0/nbinsF,P_E[i]);
 	   fprintf(OutFile, "%f %f %f\n",C_NC*1E9, j*1.0/nbinsF, P_E[j]);
         }
	 fprintf(OutFile,"\n");
        }
      }

      if (calc_N_i_C) //calculate Fraction of TAR bound as a function of NC concentration
      {
	printf("C_NC  Avg_N    Any_N  Any_N_DS\n");

        //for (i=0;i<N_i;i++)
        C_NC=0.0;
        for (i=0;i<(N_C+1);i++)
        {
          //C_NC = C_N_i[i]*1E-9;
          Any_NC_bound = Calc_N_i ( Seq, struct1, kT, &Any_NC_SS);
          Avg_N = Calc_Avg_N ( Seq, struct1, kT);
	  printf ("%g  %f  %f  %f\n",C_NC*1E9, Avg_N, Any_NC_bound, Any_NC_SS);
          C_NC += dC;
        }
      }

      if (calc_N_i)  Any_NC_bound = Calc_N_i ( Seq, struct1, kT,&Any_NC_SS);
      Avg_N = Calc_Avg_N ( Seq, struct1, kT);
      printf ("Avg_N=%f, Any_NC=%f\n",Avg_N, Any_NC_bound);


    free_pf_arrays();
    if (cstruc!=NULL) free(cstruc);
    free(base_pair);
    (void) fflush(stdout);
    free(string);
    free(structure); 
  } while (1);
  return 0;
}

PRIVATE void usage(void)
{
  nrerror("usage:\n"
	  "RNAdist [-p[0]] [-C] [-T temp] [-F force] [-4] [-d[2|3]] [-noGU] [-noCloseGU]\n" 
	  "        [-noLP] [-e e_set] [-P paramfile] [-nsp pairs] [-o ext_bases]\n"
	  "        [-v] [-b] [-S scale] [-noconv] [-f outputfile]\n");
}


/* calculate P_x from P_m */
void m_to_length(FLT_OR_DBL * P_x, FLT_OR_DBL * P_m, float xmin, float xmax, int nbins, int n)
{
/* P_x - probability of end-to-end distance x*/
/* P_m - probability of m exterior bases */
/* xmin - minimum end-to-end distance to include */
/* xmax - maximum end-to-end distance to include */
/* nbins - number of bins for P_x */
/* n - number of bases in RNA */

  float base_length=0.7;   /* length of one base (in nm)       */
  float pers_length=1.92;  /* persistence length (in nm)       */

  int i; /* bin number */
  int m; /* number of exterior bases */
  float norm; /* normalization factor */
  float k; /* coefficient for gaussian */

  float dx = (xmax-xmin)/nbins; /* size of x bins */
  float x; /* end-to-end distance in nm */ 

  for(i=0;i<nbins;i++) P_x[i]=0; /* intialize P_x */

  for (m=1;m<=n;m++)
  {
	k = 1.5/m/base_length/pers_length;
	norm = pow(k/M_PI,1.5)*dx;
	
	for(i=0;i<nbins;i++)	
	{
	  x = xmin + i*dx; /* end-to-end distane */
	  P_x[i] += P_m[m] * norm * exp(-k*x*x) * 4*M_PI*x*x; /* assume Gaussian polymer */
	}
  }  
  if (xmin==0) P_x[0] += P_m[0]; /* include zero length polymer */

}
/* calculate P_x from P_m */
void length_to_FRET(FLT_OR_DBL * P_x, FLT_OR_DBL * P_E, int nbins, float xmin, float dx, int nbinsx)
{
/* P_x - probability of end-to-end distance x*/
/* P_m - probability of m exterior bases */
/* xmin - minimum end-to-end distance to include */
/* xmax - maximum end-to-end distance to include */
/* nbins - number of bins for P_x */
/* n - number of bases in RNA */


  int i; /* bin number */
  int xi; /* number of exterior bases */
  float norm; /* normalization factor */
  float k; /* coefficient for gaussian */
  float E;
  float R;
  float R0 = R_FRET;

  float x; /* end-to-end distance in nm */ 

  for(i=0;i<nbins;i++) P_E[i]=0; /* intialize P_E */

  xi = nbinsx;
  x = dx*nbinsx+xmin;

  for (i=1;i<=nbins;i++)
  {
	E = 1.0 / nbins * i;
	R = R0 * pow((1.0/E -1.0),1.0/6.0);
	
	while ( x >= R )	
	{
	  x = xmin + xi*dx; /* end-to-end distane */
	  P_E[i] += P_x[xi];
	  xi--;
	}
  }  


}
/* calculate P_x from P_m */
float m_to_FRET(FLT_OR_DBL * P_E, FLT_OR_DBL * P_m, FLT_OR_DBL * E_m, float xmin, float xmax, int nbins, int n)
{
/* P_x - probability of end-to-end distance x*/
/* P_m - probability of m exterior bases */
/* xmin - minimum end-to-end distance to include */
/* xmax - maximum end-to-end distance to include */
/* nbins - number of bins for P_x */
/* n - number of bases in RNA */

  float base_length=0.7;   /* length of one base (in nm)       */
  float pers_length=1.92;  /* persistence length (in nm)       */
  float R0 = R_FRET;
  float R0_6 = pow(R0,6);

  int i; /* bin number */
  int m; /* number of exterior bases */
  float norm; /* normalization factor */
  float k; /* coefficient for gaussian */

  float dx = (xmax-xmin)/nbins; /* size of x bins */
  float x; /* end-to-end distance in nm */ 
  float P; /* probability of position x with m exterior bases */
  float E; /* probability FRET efficiency E with position x*/
  float Eavg;
  int Eindex;
  float Etotal;

  char Line[10000]; /*string to hold line of data from file*/
  int status=0;

  for(i=0;i<(nbins+1);i++) P_E[i]=0; /* intialize P_x */

  //printf("Opening m_to_FRET...\n");

  FILE * InputFile = fopen("m_to_FRET.dat", "r"); /*open input file*/
  if (InputFile == NULL) 
  {
  	printf("Input file not found!! Cannot compute FRET values...\n");
	return(-1.0);	  

  } 

  //printf("Reading m_to_FRET...\n");


  Etotal = 0.0;
  for (m=3;m<=n;m++)
  {
        if (fgets(Line,10000, InputFile)==NULL)
        {
  	  printf("Failed to read input file!! Cannot compute FRET values...\n");
	  return(-1.0);
        } 
	sscanf(Line,"%f", &Eavg);      

        //printf("E[%d]=%f\n",m,Eavg);

	Eindex = (int) (Eavg * nbins);
	//printf("Eavg[%d]=%f, %d\n",m,Eavg, Eindex);
	P_E[Eindex] += P_m[m];
	Etotal += Eavg * P_m[m];
	E_m[m] = Eavg;
  }  
  //printf ("Average E = %f\n", Etotal);
  fclose(InputFile);
  return(Etotal);

}

void print_NC_energies ()
{
	FLT_OR_DBL kT = ((temperature+K0)*GASCONST/1000.0);
	FLT_OR_DBL NC_Energy_ext = -0.0 / kT; /*energy for NC binding*/
	FLT_OR_DBL NC_Energy_loop = -0.0 / kT; /*energy for NC binding*/
	footprint = 6; /*minimum size for NC binding site*/
	FLT_OR_DBL C_NC = 1.0; /*concentration NC in M*/

	printf("T=%4.1f C, kT=%f kcal/mol\n",temperature, kT);
	printf("salt=%f M\n",salt);

	char *InputFileName = "NC_Energies.dat";
	char Line[10000]; /*string to hold line of data from file*/
	int status=0;
  	FILE * InputFile = fopen(InputFileName, "r"); /*open input file*/
  	if (InputFile == NULL) 
   	{
  	  printf("Input file not found!! Will use default values...\n");
	  printf("NC_Energy_ext=%f kcal/mol\n",NC_Energy_ext);
	  printf("NC_Energy_loop=%f kcal/mol\n",NC_Energy_loop);
	  printf("footprint=%d nt\n",footprint);
	  printf("C_NC=%g M\n",C_NC);
   	} else {/* Read NC energy parameters from file NC_Energies.dat */

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &NC_Energy_ext);
   	  } else status = -1;
  	  printf("NC_Energy_ext=%f kcal/mol\n",NC_Energy_ext);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &NC_Energy_loop);
   	  } else status = -1;
  	  printf("NC_Energy_loop=%f kcal/mol\n",NC_Energy_loop);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%d", &footprint);
   	  } else status = -1;
  	  printf("footprint=%d nt\n",footprint);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &C_NC);
   	  } else status = -1;
  	  printf("C_NC=%g M\n",C_NC);
	}/* End Read NC energy parameters from file NC_Energies.dat */
	fclose(InputFile);
}


double Calc_Avg_N ( char * Seq, char * struct1 , double kT)
{
    double Avg_N;
    double Energy0, Energy;
    double dmu = 0.01;
    double dC = exp(-dmu/kT); //concentration step scale
    double temp_C_NC;

    Energy0=pf_fold(Seq ,struct1);

    temp_C_NC = C_NC;

    C_NC *= dC;

    Energy=pf_fold(Seq ,struct1);

    Avg_N = (Energy-Energy0)/dmu;

    //printf("Avg_N=%f, Energy=%f, Energy0=%f, C_NC=%g, dC=%f\n",Avg_N, Energy, Energy0, C_NC, dC);

    C_NC = temp_C_NC;

    return(Avg_N);
}



double Calc_N_i ( char * Seq, char * struct1 , double kT, double*Any_NC_SS)
{
    double No_NC_bound=1.0;
    double No_NC_bound_SS=1.0;
    double * N_i;
    double N_i_conv=0.;  //convolution of N_i over footprint to show prob that each nt is bound by protein
    double Energy0, Energy;
    double dmu = 0.01;
    int i,j;
    double temp_exp_dmu;

    int n = (int) strlen(Seq); //number of bases

    N_i = (double*) malloc (n * sizeof(double));

    double N_DS;
    double * N_DS_i = (double*) malloc ( n * sizeof(double));

    FILE * N_i_C_File;
    if (print_N_i_C_dist)  //open N_i_C_File
    {
	N_i_C_File = fopen(N_i_C_FileName, "a");
	//fprintf(N_i_C_File, "%g ", C_NC*1E9);
    }


    Energy0=pf_fold(Seq ,struct1);

    for (i=1;i<=n;i++)
    {
      //binding to ssDNA
      temp_exp_dmu = exp_dmu[i];
      exp_dmu[i]*=exp(-dmu/kT);
      Energy=pf_fold(Seq ,struct1);
      N_i[i] = (Energy-Energy0)/dmu;
      exp_dmu[i]=temp_exp_dmu;
      No_NC_bound_SS *= (1.0-N_i[i]);

      //binding to dsDNA
      temp_exp_dmu = exp_dmuDS[i];
      exp_dmuDS[i]*=exp(-dmu/kT/footprintDS); 
      Energy=pf_fold(Seq ,struct1);
      //N_DS_i[i]=(Energy-Energy0)/dmu;
      N_i[i] += (Energy-Energy0)/dmu;
      /*if (i>=(footprint-1)) 
      {
	N_DS = 1.0;
	for (j=i-footprint+1;j<=i;j++) N_DS *= N_DS_i[j];
	N_i += N_DS;
      }*/
      exp_dmuDS[i]=temp_exp_dmu;

      No_NC_bound *= (1.0-N_i[i]);


      N_i_conv+=N_i[i];
      if (i>=(footprint)) N_i_conv-=N_i[i-footprint];
 
      //printf("N[ %d ] = %f %f %f %f\n",i,N_i-N_DS, N_DS, N_DS_i[i], N_i);
      //if (print_N_i_C_dist) fprintf(N_i_C_File, "%g %f %f\n",C_NC*1E9, ((double)i), N_i[i]);
      if (print_N_i_C_dist){
	 fprintf(N_i_C_File, "%g %f %f\n",C_NC*1E9, ((double)i)-0.9999, N_i_conv);
	 fprintf(N_i_C_File, "%g %f %f\n",C_NC*1E9, ((double)i), N_i_conv);
      }
      //for (j=(i-footprint+1);j<=i;j++) if (j>=0) exp_dmu[j]=1.0;
      //printf("i=%d\n",i);
    }
    /*while (i<=(n+footprint))
    {
      //printf("i=%d\n",i);
      N_i_conv-=N_i[i-footprint];
      if (print_N_i_C_dist) fprintf(N_i_C_File, "%g %f %f\n",C_NC*1E9, ((double)i), N_i_conv);
      i++;
    }*/

    if (print_N_i_C_dist) fprintf(N_i_C_File,"\n");
    *Any_NC_SS = 1-No_NC_bound_SS;
 
    if (print_N_i_C_dist)  //close N_i_C_File
    {
	fclose(N_i_C_File);
    }

    //free(N_i);
    return(1-No_NC_bound);
}

/* calculate P_x from P_m */
float m_to_FRET_dist(FLT_OR_DBL * P_E, FLT_OR_DBL * P_m, FLT_OR_DBL * E_m, float xmin, float xmax, int nbins, int n)
{
/* P_x - probability of end-to-end distance x*/
/* P_m - probability of m exterior bases */
/* xmin - minimum end-to-end distance to include */
/* xmax - maximum end-to-end distance to include */
/* nbins - number of bins for P_x */
/* n - number of bases in RNA */

  float base_length=0.7;   /* length of one base (in nm)       */
  float pers_length=1.92;  /* persistence length (in nm)       */
  float R0 = R_FRET;
  float R0_6 = pow(R0,6);

  int i; /* bin number */
  int m; /* number of exterior bases */
  float norm; /* normalization factor */
  float k; /* coefficient for gaussian */

  float dx = (xmax-xmin)/nbins; /* size of x bins */
  float x; /* end-to-end distance in nm */ 
  float P; /* probability of position x with m exterior bases */
  float E; /* probability FRET efficiency E with position x*/
  float Eavg;
  int Eindex;
  float Etotal;
  float FRET_dist;
  float E_FRET;
  int FRET_bins = 50;


  char Line[10000]; /*string to hold line of data from file*/
  int status=0;

  for(i=0;i<(nbins+1);i++) P_E[i]=0; /* intialize P_x */



  //printf("Opening m_to_FRET...\n");

  //FILE * InputFile = fopen("m_to_FRET_dist_d0,55_r0,0_100k_b2,0.txt", "r"); /*open input file*/
  FILE * InputFile = fopen("m_to_FRET_dist.txt", "r"); /*open input file*/
  if (InputFile == NULL) 
  {
  	printf("Input file not found!! Cannot compute FRET values...\n");
	return(-1.0);	  

  } 

  //printf("Reading m_to_FRET_dist...\n");


  Etotal = 0.0;
  (fgets(Line,10000, InputFile)); //read off header
  for (m=3;m<=n;m++)
  {
	(fgets(Line,10000, InputFile));
        //printf("%s\n",Line); 

        E_m[m] = 0.;
        for (i=0;i<FRET_bins+1;i++)
	{
          if (fgets(Line,10000, InputFile)==NULL)
          {
  	    printf("Failed to read input file!! Cannot compute FRET values...\n");
	    return(-1.0);
          } 
	  //printf("%s\n",Line);
	  sscanf(Line,"%f %f", &E_FRET, &FRET_dist);   

          //printf("E[%d]=%f\n",m,Eavg);

	  Eindex = (int) (E_FRET * nbins + 0.1);
	  //printf("Eavg[%d]=%f, %f, %d, %f\n",m,E_FRET * nbins, E_FRET, Eindex, FRET_dist);
	  P_E[Eindex] += P_m[m]*FRET_dist;
	  Etotal += E_FRET * P_m[m] *FRET_dist;
	  E_m[m] += E_FRET *FRET_dist;

	}   


  }  
  //printf ("Average E = %f\n", Etotal);
  fclose(InputFile);
  return(Etotal);

}

