/* Last changed Time-stamp: <2000-10-10 11:24:26 ivo> */
/* support for pulling force added <2000-05-30 rbund> */
/* support for single-stranded binding proteins added <2009-09-08 rforties> */
/*
       global variables to change behaviour of folding routines
			  Vienna RNA package
*/
#include <string.h>
#include <stdio.h>
#include "fold_vars.h"

/* parameters for nucleocapsid binding */
int footprint = 6; /* binding site size (nucleotides) */
double NC_Energy_ext = 0.0; /* binding energy for external bases (kcal/mol) */
double NC_Energy_loop = 0.0; /* binding energy inside loops (kcal/mol)*/
double R_FRET = 6.0; /* FRET radius for Cy3-Cy5 (nm)*/
double C_NC = 1.0; /* Nucleocapsid concentration (M) */
int High_NC_pos = 11; /* end position of high affinity NC binding site */
double High_NC_energy = -3.; /* extra binding energy for high affinity NC binding site (kcal/mol) */
double exp_E_NC_DS = 0; /*free energy for NC binding to dsDNA*/
double footprintDS = 6; /*footprint for NC binding to dsDNA*/

//FLT_OR_DBL *pNCr;         /* probability NC is bound at each base */

double * exp_dmu; 	/*deviation from average binding energy for NC bound starting at each base*/
double * exp_dmuDS; 	/*deviation from average binding energy for NC bound starting at each base*/

int  noGU = 0;           /* GU not allowed at all */
int  no_closingGU = 0;   /* GU allowed only inside stacks */
int  tetra_loop = 1;     /* Fold with specially stable 4-loops */
int  energy_set = 0;     /* 0 = BP; 1=any with GC; 2=any with AU parameters */
int  dangles = 1;	 /* use dangling end energies */
char *nonstandards = (char *)0; /* contains allowed non standard bases */
double temperature = 37.0;
float force = 0.0;
float salt  = 1.0;       /* monovalent salt concentration in M */
int extstemenergy = 0;   /* by default no extra energy for exterior stems */
int  james_rule = 1;     /* interior loops of size 2 get energy 0.8Kcal and
			    no mismatches (no longer used) */
struct bond  *base_pair;

FLT_OR_DBL *pr;           /* base pairing prob. matrix */
int  *iindx;              /* pr[i,j] -> pr[iindx[i]-j] */
FLT_OR_DBL *fx;           /* fixed extension free energy vector */
FLT_OR_DBL *fmn=0;        /* fixed extension free energy matrix */
FLT_OR_DBL *fmnb=0;       /* backward fixed extension free energy matrix */
FLT_OR_DBL *fl=0;         /* free energy at different length */
double pf_scale=- 1;      /* scaling factor to avoid floting point overflows */
int   fold_constrained = 0; /* guess what */
int   do_backtrack=1;     /* calculate pair prob matrix in part_func() */
int   do_fixedextension=0;/* calculate fixed ext. weight in part_func()*/
int   do_autoscale=0;     /* determine pf_scale automatically          */
int   open_bases=-1;      /* number of open bases in backtracking or -1*/
int   subtract_one=0;     /* subtract a one from all Boltzmann factors */
int   do_polymer=0;       /* emulate polymer contribution to force     */
int   return_alllengths=1;/* return all restricted partition functions */
int   noLonelyPairs = 0;  /* avoid helices of length 1 */
char backtrack_type='F';  /* 'C' require (1,N) to be bonded;
			     'M' seq is part of s multi loop */

char * option_string(void) {
  static char options[100];
  *options = '\0';
  if (noGU) strcat(options, "-noGU ");
  if (no_closingGU) strcat(options, "-noCloseGU ");
  if (!tetra_loop) strcat(options, "-4 ");
  if (noLonelyPairs) strcat(options, "-noLP ");
  if (fold_constrained) strcat(options, "-C ");
  if (dangles!=1) sprintf(options+strlen(options), "-d%d ", dangles);
  if (subtract_one) strcat(options, "-v ");
  if (do_polymer) strcat(options, "-b ");
  if (open_bases>=0) 
    sprintf(options+strlen(options), "-o %d ", open_bases);
  if (temperature!=37.0) 
    sprintf(options+strlen(options), "-T %f ", temperature);
  if (force!=0.0)
    sprintf(options+strlen(options), "-F %f ", force);
  return options;
}
