/* Last changed Time-stamp: <2000-10-10 18:05:48 ivo> */
/* support for pulling force added <2000-05-30 rbund> */
/* support for single-stranded binding proteins added <2009-09-08 rforties> */
/*                
		  partiton function for RNA secondary structures

		  Ivo L Hofacker
		  Vienna RNA package
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>    /* #defines FLT_MIN */
#include "utils.h"
#include "energy_par.h"
#include "fold_vars.h"
#include "pair_mat.h"
/*@unused@*/
static char rcsid[] = "$Id: part_func.c,v 1.3 2003/12/17 13:23:03 rbund Exp $";

#define MAX(x,y) (((x)>(y)) ? (x) : (y))
#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#define PUBLIC
#define PRIVATE static
#define STACK_BULGE1  1   /* stacking energies for bulges of size 1 */
#define NEW_NINIO     1   /* new asymetry penalty */


PUBLIC  float pf_fold(char *sequence, char *structure);
PUBLIC  void  init_pf_fold(int length);
PUBLIC  void  free_pf_arrays(void);
PUBLIC  void  update_pf_params(int length);
PUBLIC  char  bppm_symbol(float *x);
PRIVATE void restricted_pf_fold(int n, FLT_OR_DBL *qr, int upper, int Delta,
				short *S, short *S1);
PRIVATE void  sprintf_bppm(int length, char *structure);
PRIVATE void  scale_pf_params(unsigned int length);
PRIVATE void  get_arrays(unsigned int length);
PRIVATE double expLoopEnergy(int u1, int u2, int type, int type2,
			     short si1, short sj1, short sp1, short sq1);
PRIVATE void make_ptypes(const short *S, const char *structure);
PRIVATE void Read_NC_Data();

/*caclulate contribution to partition function from interaction with nucleocapsid protein (NC)*/
PRIVATE void calc_NC_int(char *sequence, FLT_OR_DBL * qNC_ext,  FLT_OR_DBL * qNC_loop); 
/* char *sequence - input nucleotide sequence, 1-D char array */
/* float *qNC_ext - qNC_ext[iindx[i]-j] gives contribution to partition function from NC binding 
to an exterior sequence of open bases starting at base i and ending at j*/
/* float *qNC_loop - qNC_ext[iindx[i]-j] gives contribution to partition function from NC binding 
to a sequeonce of open bases inside a loop starting at base i and ending at j*/

PRIVATE FLT_OR_DBL expMLclosing, expMLintern[NBPAIRS+1], *expMLbase;
PRIVATE FLT_OR_DBL expOUTintern[NBPAIRS+1], *expOUTbase, expOneOUTbase;
PRIVATE FLT_OR_DBL expTermAU;
PRIVATE FLT_OR_DBL expdangle5[NBPAIRS+1][5], expdangle3[NBPAIRS+1][5];
PRIVATE FLT_OR_DBL lxc, exptetra[40], expTriloop[40];
PRIVATE FLT_OR_DBL expstack[NBPAIRS+1][NBPAIRS+1];
PRIVATE FLT_OR_DBL expmismatchI[NBPAIRS+1][5][5],
  expmismatchH[NBPAIRS+1][5][5], expmismatchM[NBPAIRS+1][5][5];
PRIVATE FLT_OR_DBL expint11[NBPAIRS+1][NBPAIRS+1][5][5];
PRIVATE FLT_OR_DBL expint21[NBPAIRS+1][NBPAIRS+1][5][5][5];
PRIVATE FLT_OR_DBL expint22[NBPAIRS+1][NBPAIRS+1][5][5][5][5];
PRIVATE FLT_OR_DBL *exphairpin;
PRIVATE FLT_OR_DBL expbulge[MAXLOOP+1];
PRIVATE FLT_OR_DBL expinternal[MAXLOOP+1];
PRIVATE FLT_OR_DBL expninio[5][MAXLOOP+1];
PRIVATE FLT_OR_DBL *q, /* unrestricted partition function for segment [i,j], 
	also reused by Ralf as the weight that a strand of length j
	has extension i  */ 
	*qb, /*partition function for segment [i,j] with i bound to j*/
	*qm, /*multiloop contribution for segment [i,j] with at least on stem on [i,j] */
	*qqm,/*multiloop contribution for segment [i,j] with i bound to x and (x,j] unpaired */
 	*qqm1, /*qqm1[i] = qqm[i] for j=j-1 */
	*qq, /*partition function for segment [i,j] with i bound to x and (x,j] unpaired */
	*qq1; /*qq1[i] = qq[i] for j=j-1*/
PRIVATE FLT_OR_DBL *prml, *prm_l, *prm_l1, *q1k, *qln;
PRIVATE FLT_OR_DBL *scale; /* scaling factor (exponent?) to keep partition function finite */
PRIVATE char *ptype; /* precomputed array of pair types */ 
PRIVATE int *jindx;
PRIVATE int init_length; /* length in last call to init_pf_fold() */
PRIVATE FLT_OR_DBL * qNC_ext; /*qNC_ext[iindx[i]-j] gives contribution to partition function from NC binding 
to an exterior sequence of open bases starting at base i and ending at j*/
PRIVATE FLT_OR_DBL * qNC_loop; /*qNC_loop[iindx[i]-j] gives contribution to partition function from NC binding 
to a sequence of open bases within a loop starting at base i and ending at j*/
#define ISOLATED  256.0

/*-----------------------------------------------------------------*/
PUBLIC float pf_fold(char *sequence, char *structure)
{
  short *S, *S1;
  int n, i,j,k,l, ij, kl, ik, u,u1,d,ii,ll, type, type_2, tt, ov=0;
  FLT_OR_DBL temp, Q, Qmax=0, prm_MLb;
  FLT_OR_DBL prmt,prmt1;
  FLT_OR_DBL qbt1, *tmp;
  FLT_OR_DBL maxq;
  FLT_OR_DBL newlq;
  FLT_OR_DBL qboundary;
   
  float free_energy;
  int Delta;   /* number of bases which have same size as a helix */

  n = (int) strlen(sequence);
  if (n>init_length) init_pf_fold(n);  /* (re)allocate space */

  /*caclulate contribution to partition function from interaction with NC*/
  qNC_ext = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));
  qNC_loop = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));	
  calc_NC_int(sequence, qNC_ext, qNC_loop);

  S = (short *) space(sizeof(short)*(n+1));
  S1= (short *) space(sizeof(short)*(n+1));
  S[0] = n;
  for (l=1; l<=n; l++) {
    S[l]  = (short) encode_char(toupper(sequence[l-1]));
    S1[l] = alias[S[l]];
  }
  make_ptypes(S, structure);

  /* find boundary for automatic rescaling */
  if (sizeof(FLT_OR_DBL)==sizeof(float)) {
    qboundary=log(1.0e15);
  } else {
    qboundary=log(1.0e150);
  }
  
  /*array initialization ; qb,qm,q
    qb,qm,q (i,j) are stored as ((n+1-i)*(n-i) div 2 + n+1-j */

  for (d=0; d<=TURN; d++) 
    for (i=1; i<=n-d; i++) {
      j=i+d;
      ij = iindx[i]-j;
      q[ij]=1.0*scale[d+1];
      qb[ij]=qm[ij]=0.0;
    }

  for (i=1; i<=n; i++) 
    qq[i]=qq1[i]=qqm[i]=qqm1[i]=prm_l[i]=prm_l1[i]=prml[i]=0;

 
  for (j=TURN+2;j<=n; j++) { /*loop over last base (j)*/
    maxq=0.0;
    /***********************************************************************/
    /*             Construct Partition Function                            */
    /***********************************************************************/
    for (i=j-TURN-1; i>=1; i--) { /*loop over first base (i)*/
      /* construction of partition function of segment i,j*/
      /*firstly that given i bound to j : qb(i,j) */
      u = j-i-1; ij = iindx[i]-j;
      type = ptype[ij];
      if (type!=0) { /*(is base pair [ij] allowed?)*/

	/*hairpin contribution*/
	if (((type==3)||(type==4))&&no_closingGU) qbt1 = 0;
	else {
	  qbt1 = exphairpin[u];
	  qbt1 *= qNC_loop[iindx[i+1]-(j-1)];/* contribution from NC binding of open bases*/
	  if ((tetra_loop)&&(u==4)) {
	    char tl[7]={0}, *ts;
	    strncpy(tl, sequence+i-1, 6);
	    if ((ts=strstr(Tetraloops, tl)))
	      qbt1 *= exptetra[(ts-Tetraloops)/7];
	  } 
	  if (u==3) {
	    char tl[6]={0,0,0,0,0,0}, *ts;
	    strncpy(tl, sequence+i-1, 5);
	    if ((ts=strstr(Triloops, tl))) 
	      qbt1 *= expTriloop[(ts-Triloops)/6];
	    if (type>2) 
	      qbt1 *= expTermAU;
	  }
	  else /* no mismatches for tri-loops */
	    qbt1 *= expmismatchH[type][S1[i+1]][S1[j-1]];
	  
	  qbt1 *= scale[u+2];
	}/*end hairpin contribution*/

	/* interior loops with interior pair k,l */
	for (k=i+1; k<=MIN(i+MAXLOOP+1,j-TURN-2); k++) {
	  u1 = k-i-1;
	  for (l=MAX(k+TURN+1,j-1-MAXLOOP+u1); l<=j-1; l++) {
	    type_2 = ptype[iindx[k]-l];
	    if (type_2) {
	      type_2 = rtype[type_2];
	      temp = qb[iindx[k]-l] * 
		expLoopEnergy(u1, j-l-1, type, type_2,
			      S1[i+1], S1[j-1], S1[k-1], S1[l+1]);
	      /* contributions from NC binding of open bases inside loops */
	      if ((i+1)<=(k-1)) temp *= qNC_loop[iindx[i+1]-(k-1)];
	      if ((l+1)<=(j-1)) temp *= qNC_loop[iindx[l+1]-(j-1)];
	      if ((u1==0) && ((j-l-1)==0)) //if the base pairing is a stack
		temp *= 1 + exp_E_NC_DS*pow(C_NC,1.0/footprintDS)*exp_dmuDS[k]*exp_dmuDS[l];
	      qbt1 += temp;
	    }
	  }
	}/*end interior loops with interior pair k,l */

	/*multiple stem loop contribution*/
	ii = iindx[i+1]; /* ii-k=[i+1,k-1] */
	temp = 0.0;
	for (k=i+2; k<=j-1; k++) temp += qm[ii-(k-1)]*qqm1[k]; 
	tt = rtype[type];
	qbt1 += temp*expMLclosing*expMLintern[tt]*scale[2]*
	  expdangle3[tt][S1[i+1]]*expdangle5[tt][S1[j-1]];
	/*end multiple stem loop contribution*/

	/*update qb(i,j) from temporary storage qbt1*/ 
	qb[ij] = qbt1;
	if (qbt1>0.0 && j-i>100) {
	  newlq=log(qbt1)/(j-i+1);
	  if (newlq>maxq) maxq=newlq;
	}
      } /* end if (type!=0) (is base pair [ij] allowed?) */
      else qb[ij] = 0.0;

      /* construction of qqm matrix containing final stem
	 contributions to multiple loop partition function
	 from segment i,j */  
      /* qqm[i] requires i paired to unspecified base k, where i<k<j, and (k,j] is unpaired */    	 
      qqm[i]=0.0;
      for(k=i+TURN; k<=j; k++){
        ik = iindx[i]-k;  /* ii-k=[i,k] ?ii-k=[i,k]?*/
	type_2=ptype[ik]; /* type of base pairing for i and k*/
	qbt1 = qb[ik]*expMLintern[type_2]*expMLbase[j-k];/* i and k paired, (k,j] unpaired */
	if (k<j) qbt1*=qNC_loop[iindx[k+1]-j];/* contribution from NC binding of open bases*/
      	if (type_2) {/*if pairing between i and k is possible*/
	  if (i>1) qbt1 *= expdangle5[type_2][S1[i-1]];
	  if (j<n) qbt1 *= expdangle3[type_2][S1[k+1]];
	  else if (type_2>2) qbt1 *= expTermAU;
     	}
	qqm[i] += qbt1;
      }	

      /* old calculation of qqm[i]*/
      /*qqm[i] = qqm1[i]*expMLbase[1]; 
      if (type) {
	qbt1 = qb[ij]*expMLintern[type];
	if (i>1) qbt1 *= expdangle5[type][S1[i-1]];
	if (j<n) qbt1 *= expdangle3[type][S1[j+1]];
	else if (type>2) qbt1 *= expTermAU;
	qqm[i] += qbt1;
      }*/
      

      /*construction of qm matrix containing multiple loop
	partition function contributions from segment i,j */
      /* qm[ij] has at least one stem on [i,j], otherwise unrestricted */
      temp = 0.0;
      ii = iindx[i];  /* ii-k=[i,k-1] */
      for (k=i+1; k<=j; k++) temp += (qm[ii-(k-1)]+expMLbase[k-i]*qNC_loop[iindx[i]-(k-1)])*qqm[k];
      qm[ij] = (temp + qqm[i]);
      if (qm[ij]>0.0 && j-i>100) {
	newlq=log(qm[ij])/(j-i+1);
	if (newlq>maxq) maxq=newlq;
      }
      
      /*auxiliary matrix qq for cubic order q calculation below */
      /* qq[i] requires i paired to unspecified base k, where i<k<j, and (k,j] is unpaired */ 
      qq[i] = 0.0;
      for(k=i; k<=j; k++){
        ik = iindx[i]-k;  /* ii-k=[i,k] ?ii-k=[i,k]?*/
	type_2=ptype[ik]; /* type of base pairing for i and k*/
	qbt1 = qb[ik]*expOUTintern[type_2]*expOUTbase[j-k];/* i and k paired, (k,j) unpaired */
	if (k<j) qbt1 *= qNC_ext[iindx[k+1]-j];/*NC interaction with unpaired segment [k+1,j]*/
        if (type_2) { /*if pairing between i and k is possible*/ 
 	  if (i>1) qbt1 *= expdangle5[type_2][S1[i-1]];
	  if (k<n) qbt1 *= expdangle3[type_2][S1[k+1]];
        }
	qq[i] += qbt1;
      }
      /* old calculateion of qq[i] */
      /*qbt1 = qb[ij]*expOUTintern[type];
      if (type) {
	if (i>1) qbt1 *= expdangle5[type][S1[i-1]];
	if (j<n) qbt1 *= expdangle3[type][S1[j+1]];
      }
      qq[i] = qq1[i]*expOUTbase[1] + qbt1;*/

      /*construction of partition function for segment i,j */
      temp = 1.0*expOUTbase[1+j-i]*qNC_ext[ij] + qq[i];
      for (k=i; k<=j-1; k++) temp += q[ii-k]*qq[k+1];
      q[ij] = temp;
      if (temp>0.0 && j-i>100) {
	newlq=log(temp)/(j-i+1);
	if (newlq>maxq) maxq=newlq;
      }

#ifndef LARGE_PF
	 if (temp>Qmax) {
	    Qmax = temp;
	    if (Qmax>FLT_MAX/10.)
	       fprintf(stderr, "%d %d %g\n", i,j,temp);
	 }
	 if (temp>FLT_MAX) {
	   PRIVATE char msg[128];
	    sprintf(msg, "overflow in pf_fold while calculating q[%d,%d]\n"
		    "use larger pf_scale", i,j);
	    nrerror(msg);
	 }
#endif
    }/*end of loop over first base (i) to construct partition function*/

    tmp = qq1;  qq1 =qq;  qq =tmp; /*set qq1(i) to qq(i,j-1)*/
    tmp = qqm1; qqm1=qqm; qqm=tmp; /*set qqm1(i) to qqm(i,j-1)*/

    /***********************************************************************/
    /*             Rescale Partition Function                              */
    /***********************************************************************/
    /* check if we have to rescale everything */
    if (do_autoscale) {
	if ((maxq*j<0.25*qboundary && j>110) || maxq*j>qboundary) {
	      
	  FLT_OR_DBL Ratio;     /* ratio of pf_scales          */
	  long double RPower;   /* power of ratio of pf_scales */

	  /* rescale all the full length arrays */
	  if (maxq*j>qboundary) {
	    Ratio=exp(0.25*maxq);
	  } else {
	    Ratio=exp(-(0.5*qboundary/j-maxq));
	  }
	  expOneOUTbase/=Ratio;
	  scale[1]/=Ratio;
	  expMLbase[1]/=Ratio;
	  for(i=1;i<=n;i++) {
	    expOUTbase[i]=expOUTbase[i-1]*expOneOUTbase;
	    scale[i]=scale[i-1]*scale[1];
	    if (i<n) expMLbase[i]=expMLbase[i-1]*expMLbase[1];
	  }

	  /* rescale the two dimensional arrays */
	  RPower=Ratio;
	  for(i=0;i<=j;i++) {
	    for(k=1;k+i<=j;k++) {
	      q[iindx[k]-(k+i)]/=RPower;
	      qm[iindx[k]-(k+i)]/=RPower;
	      qb[iindx[k]-(k+i)]/=RPower;
	    }
	    RPower*=Ratio;
	  }

	  /* rescale the one dimensional arrays */
	  RPower=Ratio*Ratio;
	  for(i=1;i<j;i++) {
	    qq[j-i]/=RPower;
	    qq1[j-i]/=RPower;
	    qqm[j-i]/=RPower;
	    qqm1[j-i]/=RPower;
	    RPower*=Ratio;
	  }

	  pf_scale*=Ratio;
	  
	} /* end of rescaling necessary ? */
    } /* end of automatic rescaling requested ? */
  } /* end of loop over last base (j) */

  /***********************************************************************/
  /*                 Calculate Free Energy                               */
  /***********************************************************************/

  if (backtrack_type=='C')      Q = qb[iindx[1]-n];
  else if (backtrack_type=='M') Q = qm[iindx[1]-n];
  else Q = q[iindx[1]-n];
 
  /*try to print Z*/
  //printf("Q=%f, pf_scale=%f, Z=%f, kT=%f\n",Q, pf_scale, Q*pow(pf_scale,n), (temperature+K0)*GASCONST/1000.0);


  /* ensemble free energy in Kcal/mol */
  if (Q<=FLT_MIN) fprintf(stderr, "pf_scale too large\n");
  free_energy = (-log(Q)-n*log(pf_scale))*(temperature+K0)*GASCONST/1000.0;
  /* in case we abort because of floating point errors */ 
  if (n>1600) fprintf(stderr, "free energy = %8.2f\n", free_energy);


   /***********************************************************************/
   /*                 Restricted Partition Function                       */
   /***********************************************************************/
 
   /* we copy the parts of the array q which we might need later on */
   for (k=1; k<=n; k++) {
     q1k[k] = q[iindx[1] - k];
     qln[k] = q[iindx[k] -n];
   }
   q1k[0] = 1.0;
   qln[n+1] = 1.0;

   /* partition function restricted to a certain number of open bases */
   
   Delta=(int)(((float)OUT_intern)/((float)OUT_BASE)+0.5);
   if (Delta<0) Delta=0; /* we cannot work with Delta<0 */
   /* helices should not be bigger than their constituents */
   if (Delta>TURN+2) Delta=TURN+2;

   if (do_fixedextension || (open_bases>=0 && open_bases<=n)) {

     /* we will reuse q for our own purposes now                      */
     /* q[iindx[i]-j] is the weight that a strand of length j has     */
     /*               extension i                                     */
     /* note, that indeed q[iindx[i]-j] is valid for 0<=i<=j !        */

     fx=qqm;  /* recycle vector */

     /* this routine does all the work filling up the array q */
     restricted_pf_fold(n, q, 0, Delta, S, S1);
     
     for(i=0;i<=n;i++) {
       fx[i]=(-log(q[iindx[i]-n])-i*log(expOneOUTbase)-n*log(pf_scale))*
	 (temperature+K0)*GASCONST/1000.0;
     }

     /* check if restricted partition function for all lengthes are
	requested */
     if (return_alllengths) {

       /* we may have to allocate memory to be able to return this */
       if (!fmn) {
	 fmn = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));
       }

       if (fmn) {
	 for(i=0;i<=n;i++) {
	   for(j=0;j<=i;j++) {
	     fmn[iindx[j]-i]=(-log(q[iindx[j]-i])-(i-j)*log(pf_scale))*
	       (temperature+K0)*GASCONST/1000.0;
	   }
	 }
       }

       /* we need memory for the backwards part as well */
       
       if (!fmnb) {
	 fmnb = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));
       }

       if (fmnb) {
	 restricted_pf_fold(n, fmnb, 1, Delta, S, S1);
	 for(i=0;i<=n;i++) {
	   for(j=0;j<=i;j++) {
	     fmnb[iindx[j]-i]=(-log(fmnb[iindx[j]-i])-j*log(expOneOUTbase)
			       -i*log(pf_scale))*
	       (temperature+K0)*GASCONST/1000.0;
	   }
	 }
       }
       
     } /* end of: do we want all restricted partition functions ? */
     
   } /* end if (do_fixedextension || open_bases>=0) */

   /* in case open_bases asks for it, we replace the total free energy
      by the restricted one */
   if (open_bases>=0 && open_bases<=n) {
     if (backtrack_type=='C') {
       if (open_bases>0)
	 free_energy=-log(0.0);  /* impossible -> infinite free energy */
       /* for open_bases=0 the original assignment was good */
     } else
       free_energy=fx[open_bases];
   }
   /***********************************************************************/
   /*                   NC binding probability                            */
   /***********************************************************************/
   /* calculate probability NC is bound at each base*/
   // for (k=1; k<=n; k++) {
   //	pNCr[k] = q[iindx[1] - k] * q[iindx[k] -n] / q[iindx[1] -n] ; 
   //  }

   /***********************************************************************/
   /*                   Backtracking                                      */
   /***********************************************************************/
   /* backtracking to construct binding probabilities of pairs*/
   
   if (do_backtrack) {
      int probabilities_vanish;
     
      Qmax=0;

    for (k=1; k<=n; k++) {
      q1k[k] = q[iindx[1] - k];
      qln[k] = q[iindx[k] -n];
    }
    q1k[0] = 1.0;
    qln[n+1] = 1.0;
      
    pr = q;     /* recycling */

    probabilities_vanish=0;
      
    /* 1. exterior pair i,j and initialization of pr array */
    if (open_bases>=0 && open_bases<=n) {/*number of open bases restricted?*/

      FLT_OR_DBL *qo;   /* array for restricted partition functions j...n */

      qo=calloc(sizeof(FLT_OR_DBL),(n+1)*(n+2)/2+1);
      if (!finite(fx[n]) || !finite(fx[open_bases]) || !qo) {

	/* we cannot calculate the base pairing probabilities */
	if (qo) free((void *)qo);
	fprintf(stderr,
		"Base pairing probabilities not available for %d open bases.\n",
		open_bases);
	for(i=1;i<=n;i++)
	  for(j=i;j<=n;j++)
	    pr[iindx[i]-j]=0.0;
	probabilities_vanish=1;
	  
      } else { /* no problem with calculation of base pairing matrix */

	/* full partition function for this number of open bases */
	FLT_OR_DBL qn;

	/* as a first thing we have to calculate the restricted
	   partition function of the substrand j...n under the
	   constraint that it has 0<=i<=n-j+1 exterior bases.
	   We store this partition function in qo[iindx[i]-(n-j+1)]. */
	
	restricted_pf_fold(n, qo, 1, Delta, S, S1);

	/* we are now ready to initialize the array of the base
	   pairing probabilities. This array is identical with the
	   array used for the restricted partition functions of
	   the substrands 1...j. Note, however, that in order
	   to fill pr[iindx[i]-j] for a given j we only need the
	   restricted partition functions 1...(j-1) which have not been
	   overwritten yet. We just have to save the total partition
	   function at the given number of exterior bases */
	qn=q[iindx[open_bases]-n];
	for (j=n; j>=1; j--) {
	  for (i=j; i>=MAX(j-TURN,1); i--) pr[iindx[i]-j] = 0;
	  for (i=j-TURN-1; i>=1; i--) {
	    ij = iindx[i]-j;
	    type = ptype[ij];
	    if (type&&(qb[ij]>0.)) {
	      /* add all combinations of the left and right strand which
		 give the correct number of open bases */
	      temp=0.0;
	      for(k=MAX(0,open_bases-Delta-n+j);
		  k<=MIN(open_bases-Delta,i-1);k++) {
		if (i==1)
		  if (j==n)
		    temp+=scale[open_bases-Delta]/expOUTbase[open_bases-Delta];
		  else
		    temp+=qo[iindx[open_bases-Delta-k]-(n-j)]*
		      scale[k]/expOUTbase[k];
		else if (j==n)
		  temp+=q[iindx[k]-(i-1)]*
		    scale[open_bases-Delta-k]/expOUTbase[open_bases-Delta-k];
		else
		  temp+=q[iindx[k]-(i-1)]*
		    qo[iindx[open_bases-Delta-k]-(n-j)];
	      }
	      pr[ij] = temp/qn*expOUTintern[type]/expOUTbase[Delta];
	      if (i>1) pr[ij] *= expdangle5[type][S1[i-1]];
	      if (j<n) pr[ij] *= expdangle3[type][S1[j+1]];
	    } else
	      pr[ij] = 0;
	  }
	}
	
	/* get rid of intermediary matrix */
	free((void *)qo);
      }
      
    } else {   /* number of open bases not restricted */
	
      for (i=1; i<=n; i++) {
	for (j=i; j<=MIN(i+TURN,n); j++) pr[iindx[i]-j] = 0;
	for (j=i+TURN+1; j<=n; j++) {
	  ij = iindx[i]-j;
	  type = ptype[ij];
	  if (type&&(qb[ij]>0.)) {
	    pr[ij] = q1k[i-1]*qln[j+1]/q1k[n];
	    if (i>1) pr[ij] *= expdangle5[type][S1[i-1]]*expOUTintern[type];
	    if (j<n) pr[ij] *= expdangle3[type][S1[j+1]];
	    else if (type>2) pr[ij] *= expTermAU;
	  } else
	    pr[ij] = 0;
	}
      }

    } /* end of: number of open bases restricted ? */

    /* loop over end bases if probabilities do not vanish completely */
      
    for (l=n; l>TURN+1 && !probabilities_vanish; l--) {
      
      /* 2. bonding k,l as substem of 2:loop enclosed by i,j */
      for (k=1; k<l-TURN; k++) {
	kl = iindx[k]-l;
	type_2 = ptype[kl]; type_2 = rtype[type_2];
	if (qb[kl]==0) continue;
	
	for (i=MAX(1,k-MAXLOOP-1); i<=k-1; i++) 
	  for (j=l+1; j<=MIN(l+ MAXLOOP -k+i+2,n); j++) {
	    ij = iindx[i] - j;
	    type = ptype[ij];
	    if ((pr[ij]>0)) {
	     if ( ((k-i-1)==0) && ((j-l-1)==0) ) 
	     { //if the base pairing is a stack
	      pr[kl] += pr[ij]*expLoopEnergy(k-i-1, j-l-1, type, type_2,
					     S1[i+1], S1[j-1], S1[k-1], S1[l+1])
				*( 1 + exp_E_NC_DS*pow(C_NC,1.0/footprintDS)*exp_dmuDS[k]*exp_dmuDS[l] );
	     } else {
	      pr[kl] += pr[ij]*expLoopEnergy(k-i-1, j-l-1, type, type_2,
					     S1[i+1], S1[j-1], S1[k-1], S1[l+1]);
	     }
	    } 
	  }
      }
      /* 3. bonding k,l as substem of multi-loop enclosed by i,j */
      prm_MLb = 0.;
      if (l<n) for (k=2; k<l-TURN; k++) {
	i = k-1;
	prmt = prmt1 = 0.0;
	    
	ii = iindx[i];     /* ii-j=[i,j]     */
	ll = iindx[l+1];   /* ll-j=[l+1,j-1] */
	tt = ptype[ii-(l+1)]; tt=rtype[tt];
	prmt1 = pr[ii-(l+1)]*expMLclosing*expMLintern[tt]*
	  expdangle3[tt][S1[i+1]]*expdangle5[tt][S1[l]];
	for (j=l+2; j<=n; j++) {
	  tt = ptype[ii-j]; tt = rtype[tt];
	  prmt += pr[ii-j]*expdangle3[tt][S1[i+1]]*
	    expdangle5[tt][S1[j-1]] *qm[ll-(j-1)];
	}
	kl = iindx[k]-l;
	tt = ptype[kl];
	prmt *= expMLclosing*expMLintern[tt];
	prml[ i] = prmt;
	prm_l[i] = prm_l1[i]*expMLbase[1]+prmt1;

	prm_MLb = prm_MLb*expMLbase[1] + prml[i];
	/* same as:    prm_MLb = 0;
	   for (i=1; i<=k-1; i++) prm_MLb += prml[i]*expMLbase[k-i-1]; */

	prml[i] = prml[ i] + prm_l[i];
	    
	if (qb[kl] == 0.) continue; 
	    
	temp = prm_MLb;

	for (i=1;i<=k-2; i++) 
	  temp += prml[i]*qm[iindx[i+1] - (k-1)];

	temp *= expMLintern[tt]*scale[2];
	if (k>1) temp *= expdangle5[tt][S1[k-1]];
	if (l<n) temp *= expdangle3[tt][S1[l+1]];
	pr[kl] += temp;
#ifndef LARGE_PF
	if (pr[kl]>Qmax) {
	  Qmax = pr[kl];
	  if (Qmax>FLT_MAX/10.)
	    fprintf(stderr, "%d %d %g %g\n", i,j,pr[kl],qb[kl]);
	}
	if (pr[kl]>FLT_MAX) {
	  ov++;
	  pr[kl]=FLT_MAX;
	}
#endif
      } /* end for (k=..) */
      tmp = prm_l1; prm_l1=prm_l; prm_l=tmp;

    }  /* end for (l=..)   */
      
    for (i=1; i<=n; i++)
      for (j=i+TURN+1; j<=n; j++) {
	ij = iindx[i]-j;
	pr[ij] *= qb[ij];
      }
      
    if (structure!=NULL)
      sprintf_bppm(n, structure);
  }   /* end if (do_backtrack)*/
   
  free(S);
  free(S1);
  if (ov>0) fprintf(stderr, "%d overflows occurred while backtracking;\n"
		    "you might try a smaller pf_scale than %g\n",
		    ov, pf_scale);
  /* assign partial free energies */
  for(i=1;i<=n;i++) {
    q1k[i] = (-log(q1k[i])-i*log(pf_scale))*(temperature+K0)*GASCONST/1000.0;
  }
  fl=q1k;
  
  free(qNC_ext);
  free(qNC_loop);

  return free_energy; 
}

/*------------------------------------------------------------------------*/
/* calculates partition function restricted to a given number of exterior bases */
PRIVATE void restricted_pf_fold(int n, FLT_OR_DBL *qr, int upper, int Delta,
				short *S, short *S1)
/* n - length of RNA sequence */
/* qr - qr[mj] = q[1j] restricted to having m exterior bases*/
/* upper - if nonzero we are instead evaluating qr*[mj]=q[n+1-j,n] with m exterior bases */
/* Delta - each exterior stem counts as Delta exterior bases (default 3) */
/* S - RNA sequece */
/* S1 - duplicate copy of RNA sequence? */

{
  int i,j,k;   /* loop variables                                */
  int l,m;     /* current bases                                 */
  int type;    /* type of a base pair                           */
  FLT_OR_DBL temp,qbt1; /* temporary variables                  */
  FLT_OR_DBL *qqr; /* qr restricted to have last base closed (or first base closed for upper)	*/

  /*allocate memory for qqr*/
  qqr = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));	

  /* The partition function for the substrand 1...j with m open
     bases is stored in qr[iindx[m]-j]. The true partition function
     is expOUTbase[m]*qr[iindx[m]-j]/scale[j] */

  
  /* for lengths smaller than TURN+2 the only possible structure
     is an open strand. Thus the weights for this case are kind of
     trivial */
  for(j=1;j<TURN+2 && j<=n;j++) {
     for(i=0;i<j;i++)
     {
	qr[iindx[i]-j]=0.0;
	qqr[iindx[i]-j]=0.0;
     }
     qr[iindx[j]-j]=qNC_ext[iindx[1]-j]/*NC binding contribution for totally open substrand*/;
  }

  /* now we have to do all the other lengths */
  for(j=TURN+2;j<=n;j++) {

    /* extension 0 */
    temp=0.0;
    if (!Delta) { /* only possible for Delta=0 */
      for(k=1;k<j-TURN;k++) {
	l=upper?(n+1-j):k;
	m=upper?(n+1-k):j;
	qbt1=qb[iindx[l]-m];
	if (qbt1>0.0) {   /* are l and m a possible pair ? */
	  type=ptype[iindx[l]-m];
	  /* add dangles and piece up to k-1 (must have ext. 0 as well) */
	  if (l>1) qbt1*=expdangle5[type][S1[l-1]];
	  if (k>1) qbt1*=qr[iindx[0]-(k-1)];
	  if (m<n) qbt1*=expdangle3[type][S1[m+1]];
	  if (backtrack_type=='M')
	    qbt1*=expMLintern[type];
	  else
	    qbt1*=expOUTintern[type];
	  temp+=temp;
	} /* end of if l-m is good base pair */
      } /* end of loop over base k which is bound to base j */
    } /* end of if (!Delta) */
    qr[iindx[0]-j]=temp;
    
    /* all extensions 0<i<=j */
    for(i=1;i<=j;i++) {
      /* get contribution of free base j, i.e., strand one shorter with
	 length one smaller */
      /*temp=qr[iindx[i-1]-(j-1)];*/ /************Does not work for NC***************/
      /* apply free base corrections */
      /*if (backtrack_type=='M')
	temp*=expMLbase[1];*/

      temp = 0.0;
      
      /* now we will add all the possibilities for which base j is bound */
      if (i==Delta) { /* we have to use one single stem in this case */
	l=upper?(n+1-j):1;
	m=upper?n:j;
	qbt1=qb[iindx[l]-m]/expOUTbase[Delta];
	if (qbt1>0.0) {   /* are l and m a possible pair ? */
	  type=ptype[iindx[l]-m];
	  /* add dangles */
	  if (m<n) qbt1*=expdangle3[type][S1[m+1]];
	  if (l>1) qbt1*=expdangle5[type][S1[l-1]];
	  if (backtrack_type=='M')
	    qbt1*=expMLintern[type];
	  else
	    qbt1*=expOUTintern[type];
	  temp+=qbt1;
	}
      } else if (i>Delta) {  /* all combinations with j paired to k, plus piece up to k-1 */
	for(k=i-Delta+1;k<j-TURN;k++) {
	  l=upper?(n+1-j):k;
	  m=upper?(n+1-k):j;
	  qbt1=qb[iindx[l]-m]/expOUTbase[Delta];
	  if (qbt1>0.0) {   /* are l and m a possible pair ? */
	    type=ptype[iindx[l]-m];
	    qbt1*=qr[iindx[i-Delta]-(k-1)]; /* add piece up to k-1 */
	    /* add dangles */
	    if (l>1) qbt1*=expdangle5[type][S1[l-1]];
	    if (m<n) qbt1*=expdangle3[type][S1[m+1]];
	    if (backtrack_type=='M')
	      qbt1*=expMLintern[type];
	    else
	      qbt1*=expOUTintern[type];
	    temp+=qbt1;
	  } /* end of if l-m is good base pair */
	} /* end of loop over base k which is bound to base j */
      } /* end if (i-Delta>=0), if i<Delta qr=0 */
      qqr[iindx[i]-j]=temp; /* qr restricted to have last base closed */

      /*sum over all combinations with k through j unpaired, k-1 paired*/
      for(k=Delta+j-i;k<=j;k++)
      {
	l=upper?(n+1-j):k;
	m=upper?(n+1-k):j;
	qbt1 = qqr[iindx[i-(j-k+1)]-(k-1)]; /* piece from 1 to k-1, with k-1 paired */
	qbt1 *= qNC_ext[iindx[l]-m]; /* open strand from k to j */
	if (backtrack_type=='M')
	  qbt1*=expMLbase[m-l]; /* apply free base corrections */
	temp += qbt1;
      }
      if (i==j) /*totally open strand*/
      {
	l=upper?(n+1-j):1;
	m=upper?(n):j;
	qbt1 = qNC_ext[iindx[l]-m];
	if (backtrack_type=='M')
	  qbt1*=expMLbase[i]; /* apply free base corrections */
	temp += qbt1;
      }
      qr[iindx[i]-j]=temp;

      fflush(stdout);
    } /* end of loop over extensions i */
    
  } /* end of loop over strand length j */

  free(qqr);
}

/*------------------------------------------------------------------------*/
/* dangling ends should never be destabilizing, i.e. expdangle>=1         */
/* specific heat needs smooth function (2nd derivative)                   */
/* we use a*(sin(x+b)+1)^2, with a=2/(3*sqrt(3)), b=Pi/6-sqrt(3)/2,       */
/* in the interval b<x<sqrt(3)/2                                          */

#define SCALE 10
#define SMOOTH(X) ((X)/SCALE<-1.2283697)?0:(((X)/SCALE>0.8660254)?(X):\
          SCALE*0.38490018*(sin((X)/SCALE-0.34242663)+1)*(sin((X)/SCALE-0.34242663)+1))

PRIVATE void scale_pf_params(unsigned int length)
{
  /* scale energy parameters and pre-calculate Boltzmann weights */
  unsigned int i, j, k, l;
  double  kT, TT;
  double  GT;

  printf("force=%f\n",force); 
   
  kT = (temperature+K0)*GASCONST;   /* kT in cal/mol  */
  TT = (temperature+K0)/(Tmeasure);

   /* scaling factors (to avoid overflows) */
  if (pf_scale==-1) { /* mean energy for random sequences: 184.3*length cal */
    pf_scale = exp(-(-185+(temperature-37.)*7.27)/kT);
    if (pf_scale<1) pf_scale=1;
  }
  scale[0] = 1.;
  for (i=1; i<=length; i++) {
    scale[i] = scale[i-1]/pf_scale;
  }

  /* loop energies: hairpins, bulges, interior, mulit-loops */
  for (i=0; i<=MIN(30,length); i++) {
    GT =  hairpin37[i]*TT;
    exphairpin[i] = exp( -GT*10./kT);
  }
  for (i=0; i<=MIN(30, MAXLOOP); i++) {
    GT =  bulge37[i]*TT;
    expbulge[i] = exp( -GT*10./kT);
    GT =  internal_loop37[i]*TT;
    expinternal[i] = exp( -GT*10./kT);
  }
  /* special case of size 2 interior loops (single mismatch) */
  if (james_rule) expinternal[2] = exp ( -80*10/kT);
   
  lxc = lxc37*TT;
  for (i=31; i<length; i++) {
    GT = hairpin37[30]*TT + (lxc*log( i/30.));
    exphairpin[i] = exp( -GT*10./kT);
  }
  for (i=31; i<=MAXLOOP; i++) {
    GT = bulge37[30]*TT + (lxc*log( i/30.));
    expbulge[i] = exp( -GT*10./kT);
    GT = internal_loop37[30]*TT + (lxc*log( i/30.));
    expinternal[i] = exp( -GT*10./kT);
  }

  for (i=0; i<5; i++) {
    GT = F_ninio37[i]*TT;
    for (j=0; j<=MAXLOOP; j++)
      expninio[i][j]=exp(-MIN(MAX_NINIO,j*GT)*10/kT);
  }
  for (i=0; (i*7)<strlen(Tetraloops); i++) {
    GT = TETRA_ENTH37 - (TETRA_ENTH37-TETRA_ENERGY37[i])*TT;
    exptetra[i] = exp( -GT*10./kT);
  }
  for (i=0; (i*5)<strlen(Triloops); i++) 
    expTriloop[i] = exp(-Triloop_E37[i]*10/kT);

  GT =  ML_closing37*TT;
  expMLclosing = exp( -GT*10/kT);

  for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
    GT =  ML_intern37*TT;
    /* if (i>2) GT += TerminalAU; */
    expMLintern[i] = exp( -GT*10./kT);
  }
  expTermAU = exp(-TerminalAU*10/kT);

  GT =  ML_BASE37*TT;
  for (i=0; i<length; i++) {
    expMLbase[i] = exp( -10.*i*GT/kT)*scale[i];
  }

   /* for the force parameters we assume that the force is measured in pN
      and the distance parameters are given in 0.1A. We have to convert
      the corresponding energy into cal/mol which is what the factor
      1.4385688 is for. Positive forces correspond to pulling. */

   if (do_polymer) {
     GT = PL_intern*force*1.4385688/kT;
     if (fabs(GT)<0.01)
       GT=1.0+GT*GT/6.0;
     else
       GT=(exp(GT)-exp(-GT))/2.0/GT;
     for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
       expOUTintern[i]=pow(GT,OUT_intern/PL_intern);
       /* if (i>2) expOUTintern[i] *= exp( -TerminalAU*10./kT); */
     }
   } else {
     for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
       GT =  -OUT_intern*force*0.14385688;
       /* if (i>2) GT += TerminalAU; */
       expOUTintern[i] = exp( -GT*10./kT);
     }
   }
   for (i=0; i<=NBPAIRS; i++) { /* includes AU penalty */
     expOUTintern[i]*=exp(-extstemenergy*10.0/kT);
   }

   if (do_polymer) {
     GT = PL_BASE*force*1.4385688/kT;
     if (fabs(GT)<0.01)
       GT=1.0+GT*GT/6.0;
     else
       GT=(exp(GT)-exp(-GT))/2.0/GT;
     GT = kT*OUT_BASE/PL_BASE*log(GT)/1.4385688;
   } else
     GT =  OUT_BASE*force;
   expOneOUTbase=exp( GT*1.4385688/kT)*scale[1];
   for (i=0; i<=length; i++) {
      expOUTbase[i] = exp( i*GT*1.4385688/kT)*scale[i];
   }

  /* if dangles==0 just set their energy to 0,
      don't let dangle energies become > 0 (at large temps),
      but make sure go smoothly to 0                        */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=4; j++) {
      GT = dangle5_H[i][j] - (dangle5_H[i][j] - dangle5_37[i][j])*TT;
      expdangle5[i][j] = dangles?exp(SMOOTH(-GT)*10./kT):1.;
      GT = dangle3_H[i][j] - (dangle3_H[i][j] - dangle3_37[i][j])*TT;
      expdangle3[i][j] =  dangles?exp(SMOOTH(-GT)*10./kT):1.;
      if (i>2) /* add TermAU penalty into dangle3 */
	expdangle3[i][j] *= expTermAU;
    }

  /* stacking energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++) {
      if (subtract_one) {
	double temp;
	temp=exp(-(stack37[i][j]*10-175*log(salt))/(GASCONST*Tmeasure));
	GT =  enthalpies[i][j]*(1.0+1.0/temp) -
	  (enthalpies[i][j]*(1.0+1.0/temp)
	   +0.1*GASCONST*Tmeasure*log(1.0+temp))*TT;
	expstack[i][j] = exp( -GT*10/kT)-1.0;
      } else {
	GT =  enthalpies[i][j] -
	  (enthalpies[i][j] - stack37[i][j]+17.5*log(salt))*TT;
	expstack[i][j] = exp( -GT*10/kT);
      }
    }

  /* mismatch energies */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<5; j++)
      for (k=0; k<5; k++) {
	GT = mism_H[i][j][k] - (mism_H[i][j][k] - mismatchI37[i][j][k])*TT;
	expmismatchI[i][j][k] = exp(-GT*10.0/kT);
	GT = mism_H[i][j][k] - (mism_H[i][j][k] - mismatchH37[i][j][k])*TT;
	expmismatchH[i][j][k] = exp(-GT*10.0/kT);
	GT = mism_H[i][j][k] - (mism_H[i][j][k] - mismatchM37[i][j][k])*TT;
	expmismatchM[i][j][k] = exp(-GT*10.0/kT);
      }

  /* interior lops of length 2 */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  GT = int11_H[i][j][k][l] -
	    (int11_H[i][j][k][l] - int11_37[i][j][k][l])*TT;
	  expint11[i][j][k][l] = exp(-GT*10./kT);
	}
  /* interior 2x1 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m;
	  for (m=0; m<5; m++) {
	    GT = int21_H[i][j][k][l][m] - 
	      (int21_H[i][j][k][l][m] - int21_37[i][j][k][l][m])*TT;
	    expint21[i][j][k][l][m] = exp(-GT*10./kT);
	  }
	}
  /* interior 2x2 loops */
  for (i=0; i<=NBPAIRS; i++)
    for (j=0; j<=NBPAIRS; j++)
      for (k=0; k<5; k++)
	for (l=0; l<5; l++) {
	  int m,n;
	  for (m=0; m<5; m++)
	    for (n=0; n<5; n++) {            
	      GT = int22_H[i][j][k][l][m][n] -
		(int22_H[i][j][k][l][m][n]-int22_37[i][j][k][l][m][n])*TT;
	      expint22[i][j][k][l][m][n] = exp(-GT*10./kT);
	    }
	}  
}

/*----------------------------------------------------------------------*/
/* calculate Boltzmann weight for an interior loop, bulge or stack */
PRIVATE double expLoopEnergy(int u1, int u2, int type, int type2,
			     short si1, short sj1, short sp1, short sq1) {
/* for an interior loop with i paired to j, p paired to q,	 */
/*	 i<p<q<j, i through p unpaired, and q through j unpaired */
/*								*/
/*  u1 = p-i-1 = length of first unpaired stretch 		*/
/*  u2 = j-q-1 = length of second unpaired stretch 		*/
/*  type = type of pairing between i and j			*/
/*	1-GC, 2-CG, 3-UG, 4-GU, 5-UA, 6-AU 			*/
/*  type2 = type of pairing between p and q			*/
/*  dangling bases: si1=S1[i+1],  sj1=S1[j-1], sp1=S1[p-1], sq1=S1[q+1] */
/*	where S1[i] is the base (A,C,G or U) located at position i */

  double z=0;
  int no_close = 0;

  if ((no_closingGU) && ((type2==3)||(type2==4)||(type==2)||(type==4)))
    no_close = 1;

  if ((u1==0) && (u2==0)) /* stack */
    z = expstack[type][type2];
  else if (no_close==0) {
    if ((u1==0)||(u2==0)) { /* bulge */
      int u;
      u = (u1==0)?u2:u1;
      z = expbulge[u];
      if (u2+u1==1) z *= expstack[type][type2];
      else {
	if (type>2) z *= expTermAU;
	if (type2>2) z *= expTermAU;
      }
    }
    else {     /* interior loop */
      if (u1+u2==2) /* size 2 is special */
	z = expint11[type][type2][si1][sj1];
      else if ((u1==1) && (u2==2)) 
	z = expint21[type][type2][si1][sq1][sj1];
      else if ((u1==2) && (u2==1))
	z = expint21[type2][type][sq1][si1][sp1];
      else if ((u1==2) && (u2==2))
	z = expint22[type][type2][si1][sp1][sq1][sj1];
      else {
	z = expinternal[u1+u2]*
	  expmismatchI[type][si1][sj1]*
	  expmismatchI[type2][sq1][sp1];
	z *= expninio[2][abs(u1-u2)];
      }
    }
  }
  return z*scale[u1+u2+2];
}
 
/*----------------------------------------------------------------------*/

PRIVATE void get_arrays(unsigned int length)
{
  unsigned int size,i;
   
  size = sizeof(FLT_OR_DBL) * ((length+1)*(length+2)/2);
  q   = (FLT_OR_DBL *) space(size);
  qb  = (FLT_OR_DBL *) space(size);
  qm  = (FLT_OR_DBL *) space(size);
  ptype = (char *) space(sizeof(char)*((length+1)*(length+2)/2));
  q1k = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  qln = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  qq  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  qq1 = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  qqm  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  qqm1 = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  prm_l = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  prm_l1 =(FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  prml = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+2));
  exphairpin = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  expMLbase  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  expOUTbase  = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  scale = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  iindx = (int *) space(sizeof(int)*(length+1));
  jindx = (int *) space(sizeof(int)*(length+1));
  for (i=0; i<=length; i++) {
    iindx[i] = ((length+1-i)*(length-i))/2 +length+1;
    jindx[i] = (i*(i-1))/2;
  }
  //pNCr = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL)*(length+1));
  //printf ("Intializing exp_dmu\n");
  exp_dmu = (double *) space(sizeof(double)*(length+1));
  for (i=0;i<(length+1);i++) exp_dmu[i] = 1.; //initialize to 1
  exp_dmuDS = (double *) space(sizeof(double)*(length+1));
  for (i=0;i<(length+1);i++) exp_dmuDS[i] = 1.; //initialize to 1
  //printf ("Done Intializing exp_dmu\n");

}

/*----------------------------------------------------------------------*/
   
PUBLIC void init_pf_fold(int length)
{
  if (length<1) nrerror("init_pf_fold: length must be greater 0");
  if (init_length>0) free_pf_arrays(); /* free previous allocation */
#ifdef SUN4
  nonstandard_arithmetic();
#else
#ifdef HP9
  fpsetfastmode(1);
#endif
#endif
  make_pair_matrix();
  get_arrays((unsigned) length);
  scale_pf_params((unsigned) length);
  init_length=length;
  Read_NC_Data();

}

PUBLIC void free_pf_arrays(void)
{
  free(q);
  fflush(stdout);
  free(qb);
  free(qm);
  free(ptype);
  free(qq); free(qq1);
  free(qqm); free(qqm1);
  free(q1k); free(qln);
  free(prm_l); free(prm_l1); free(prml);
  free(exphairpin);
  free(expMLbase);
  free(expOUTbase);
  free(scale);
  free(iindx); free(jindx);
  //free(pNCr);
  free(exp_dmu);
  free(exp_dmuDS);
  if (fmn) free(fmn);
  fmn=0;
  if (fmnb) free(fmnb);
  fmnb=0;
#ifdef SUN4
  standard_arithmetic();
#else
#ifdef HP9
  fpsetfastmode(0);
#endif
#endif
  init_length=0;
}
/*---------------------------------------------------------------------------*/

PUBLIC void update_pf_params(int length)
{
  make_pair_matrix();
  scale_pf_params((unsigned) length);
}

/*---------------------------------------------------------------------------*/

PUBLIC char bppm_symbol(float *x)
{
  if( x[0] > 0.667 )  return '.';                  
  if( x[1] > 0.667 )  return '(';                  
  if( x[2] > 0.667 )  return ')';
  if( (x[1]+x[2]) > x[0] ) {
    if( (x[1]/(x[1]+x[2])) > 0.667) return '{';
    if( (x[2]/(x[1]+x[2])) > 0.667) return '}';
    else return '|';
  }
  if( x[0] > (x[1]+x[2]) ) return ',';
  return ':';
}

/*---------------------------------------------------------------------------*/
#define L 3
PRIVATE void sprintf_bppm(int length, char *structure)
{
  int    i,j;
  float  P[L];   /* P[][0] unpaired, P[][1] upstream p, P[][2] downstream p */
         
  for( j=1; j<=length; j++ ) {
    P[0] = 1.0;
    P[1] = P[2] = 0.0;
    for( i=1; i<j; i++) {
      P[2] += pr[iindx[i]-j];    /* j is paired downstream */
      P[0] -= pr[iindx[i]-j];    /* j is unpaired */
    }
    for( i=j+1; i<=length; i++ ) {
      P[1] += pr[iindx[j]-i];    /* j is paired upstream */
      P[0] -= pr[iindx[j]-i];    /* j is unpaired */
    }
    structure[j-1] = bppm_symbol(P);
  }
  structure[length] = '\0';
}   

/*---------------------------------------------------------------------------*/
PRIVATE void make_ptypes(const short *S, const char *structure) {
  int n,i,j,k,l;
  
  n=S[0];
  for (k=1; k<n-TURN-1; k++) 
    for (l=1; l<=2; l++) {
      int type,ntype=0,otype=0;
      i=k; j = i+TURN+l;
      type = pair[S[i]][S[j]];
      while ((i>=1)&&(j<=n)) {
	if ((i>1)&&(j<n)) ntype = pair[S[i-1]][S[j+1]];
	if (noLonelyPairs && (!otype) && (!ntype)) 
	  type = 0; /* i.j can only form isolated pairs */
	qb[iindx[i]-j] = 0.;
	ptype[iindx[i]-j] = (char) type;
	otype =  type;
	type  = ntype;
	i--; j++;
      }
      
    }
  
  if (fold_constrained&&(structure!=NULL)) {
    int hx, *stack;
    char type;
    stack = (int *) space(sizeof(int)*(n+1));
    
    for(hx=0, j=1; j<=n; j++) {
      switch (structure[j-1]) {
      case 'x': /* can't pair */ 
	for (l=1; l<j-TURN; l++) ptype[iindx[l]-j] = 0;
	for (l=j+TURN+1; l<=n; l++) ptype[iindx[j]-l] = 0;
	break;
      case '(':
	stack[hx++]=j;
	/* fallthrough */
      case '<': /* pairs upstream */
	for (l=1; l<j-TURN; l++) ptype[iindx[l]-j] = 0;
	break;
      case ')':
	if (hx<=0) {
	  fprintf(stderr, "%s\n", structure);
	  nrerror("unbalanced brackets in constraints");
	}
	i = stack[--hx];
	type = ptype[iindx[i]-j];
	/* don't allow pairs i<k<j<l */
	for (k=i; k<=j; k++)
	  for (l=j; l<=n; l++) ptype[iindx[k]-l] = 0;
	/* don't allow pairs k<i<l<j */
	for (k=1; k<=i; k++)
	  for (l=i; l<=j; l++) ptype[iindx[k]-l] = 0;
	ptype[iindx[i]-j] = (type==0)?7:type;
	/* fallthrough */
      case '>': /* pairs downstream */
	for (l=j+TURN+1; l<=n; l++) ptype[iindx[j]-l] = 0;
	break;
      }
    }
    if (hx!=0) {
      fprintf(stderr, "%s\n", structure);
      nrerror("unbalanced brackets in constraint string");
    }
    free(stack);
  }
}

/*caclulate contribution to partition function from interaction with nucleocapsid protein (NC)*/
PRIVATE void calc_NC_int(char *sequence, FLT_OR_DBL * qNC_ext,  FLT_OR_DBL * qNC_loop)
/* char *sequence - input nucleotide sequence, 1-D char array */
/* float *qNC_ext - qNC[iindx[i]-j] gives contribution to partition function from NC binding 
to an exterior sequence of open bases starting at base i and ending at j*/
{
	int n /*length of RNA sequence*/,  i,j,k/*loop counters*/;
	FLT_OR_DBL kT = ((temperature+K0)*GASCONST/1000.0);

	int N_subseq; /* size of subsequence from i to j */
	FLT_OR_DBL expNC_Energy_ext; /*e^(-NC_Energy_ext/kT)*/
	FLT_OR_DBL expNC_Energy_loop; /*e^(-NC_Energy_loop/kT)*/
	FLT_OR_DBL expNC_HighEnergy; /*exp(-High_NC_energy/kT)*/

      
        //number of base in sequence
	n = (int) strlen(sequence); 

        //sequence dependent chemical potential (difference from average)
        //double * exp_dmu = malloc (n*sizeof(double));
        //for (i=0;i<n;i++) exp_dmu[i] = 1.;


	expNC_Energy_ext= exp(-NC_Energy_ext/kT);
	expNC_Energy_loop=exp(-NC_Energy_loop/kT);	
	expNC_HighEnergy=exp(-High_NC_energy/kT);
		

        //printf ("Calculating expNC\n");

	for(j=1;j<=n;j++) 
	{
	   for(i=j;i>=1;i--)
	   {
		N_subseq = j-i+1;


                //Start with no NCs bound
	        qNC_ext[iindx[i]-j]=1.0;
		qNC_loop[iindx[i]-j]=1.0;
               
                if (N_subseq == footprint)
                {
                 if (i>1)
                 {
		  if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = 1.0 + expNC_Energy_ext*expNC_HighEnergy*exp_dmu[i]*C_NC;
		  if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = 1.0 + expNC_Energy_loop*expNC_HighEnergy*exp_dmu[i]*C_NC;
                 } else {
		  if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = 1.0 + expNC_Energy_ext*exp_dmu[i]*C_NC;
		  if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = 1.0 + expNC_Energy_loop*exp_dmu[i]*C_NC;
                 }
		}

                if (((N_subseq - footprint)>0) && (i>1||j<n))
                {   

                  //start with partition function for strand 1 shorter
		  if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = qNC_ext[iindx[i]-(j-1)];
		  if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = qNC_loop[iindx[i]-(j-1)];				
              
	          //last NC bound at j-footprint+1 (not next to helix)
                  k=j-footprint+1;
                  if (k>i) //!!!!This should be exp_dmu[i]?!?!?!?
                  {
		      if (NC_Energy_ext<0) qNC_ext[iindx[i]-j]+=qNC_ext[iindx[i]-(k-1)]	
						*expNC_Energy_ext*exp_dmu[k]*C_NC;
		      if (NC_Energy_loop<0) qNC_loop[iindx[i]-j]+=qNC_loop[iindx[i]-(k-1)]	
						*expNC_Energy_loop*exp_dmu[k]*C_NC;
		  } else { //!!!!This should be exp_dmu[i]?!?!?!?
		      if (NC_Energy_ext<0) qNC_ext[iindx[i]-j]+= expNC_Energy_ext*exp_dmu[k]*C_NC;
		      if (NC_Energy_loop<0) qNC_loop[iindx[i]-j]+=expNC_Energy_loop*exp_dmu[k]*C_NC;
		  }

                } 
	   }
	}  

        //printf ("Add helix binding\n");

	for(j=n;j>=1;j--) 
	{
	   for(i=j;i>=1;i--)
	   {
              k=j-footprint+1;
              if (k>i)
              { 
                  if (j<n)
                  {   //High Energy NC binding at last basepair
		      if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = qNC_ext[iindx[i]-(j-1)] 	
					+qNC_ext[iindx[i]-(k-1)]*expNC_Energy_ext*expNC_HighEnergy*exp_dmu[k]*C_NC;	
		      if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = qNC_loop[iindx[i]-(j-1)] 	
					+qNC_loop[iindx[i]-(k-1)]*expNC_Energy_loop*expNC_HighEnergy*exp_dmu[k]*C_NC; 
                  } else {
		      if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = qNC_ext[iindx[i]-(j-1)] 	
					+qNC_ext[iindx[i]-(k-1)]*expNC_Energy_ext*exp_dmu[k]*C_NC;	
		      if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = qNC_loop[iindx[i]-(j-1)] 	
					+qNC_loop[iindx[i]-(k-1)]*expNC_Energy_loop*exp_dmu[k]*C_NC;
                  }
               } else if ((k==i) && (j<n)) {
    		  if (NC_Energy_ext<0) qNC_ext[iindx[i]-j] = 1.0 + expNC_Energy_ext*expNC_HighEnergy*exp_dmu[i]*C_NC;
		  if (NC_Energy_loop<0) qNC_loop[iindx[i]-j] = 1.0 + expNC_Energy_loop*expNC_HighEnergy*exp_dmu[i]*C_NC;
               }
	   }
	}  

        /*printf("Q_NC0 = %f\n",expNC_Energy_ext*expNC_HighEnergy*C_NC);

        for (i=0;i<3;i++)
        {
           for(j=i;j<20;j++) 
                if(qNC_ext[iindx[i]-j]>1.0)
		printf("[%d,%d]=%g %g\n",i,j,qNC_ext[iindx[i]-j],qNC_loop[iindx[i]-j]);
        }*/

 
	//free(exp_dmu);

}

PRIVATE void Read_NC_Data()
{
        char Line[10000]; /*string to hold line of data from file*/
        int status=0;
	char *InputFileName = "NC_Energies.dat";
  	FILE * InputFile = fopen(InputFileName, "r"); /*open input file*/
  	if (InputFile == NULL) 
   	{
  	  printf("Input file not found!! Will use default values...\n");
	  //printf("NC_Energy_ext=%f\n",NC_Energy_ext);
	  //printf("NC_Energy_ext=%f\n",NC_Energy_loop);
	  //printf("footprint=%d\n",footprint);
	  //printf("C_NC=%g\n",C_NC);
   	} else {/* Read NC energy parameters from file NC_Energies.dat */

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &NC_Energy_ext);
   	  } else status = -1;
  	  //printf("NC_Energy_ext=%f\n",NC_Energy_ext);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &NC_Energy_loop);
   	  } else status = -1;
  	  //printf("NC_Energy_ext=%f\n",NC_Energy_loop);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%d", &footprint);
   	  } else status = -1;
  	  //printf("footprint=%d\n",footprint);

  	  if (fgets(Line,10000, InputFile)!=NULL)
   	  {
   	    sscanf(Line,"%lf", &C_NC);
   	  } else status = -1;
  	  //printf("C_NC=%g\n",C_NC);
	}/* End Read NC energy parameters from file NC_Energies.dat */
	fclose(InputFile);

}
