/* Last changed Time-stamp: <2000-10-10 18:05:48 ivo> */
/* support for pulling force added <2000-05-30 rbund> */
/*                
		  partiton function for RNA secondary structures

		  Ivo L Hofacker
		  Vienna RNA package
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>    /* #defines FLT_MIN */
#include "utils.h"
#include "energy_par.h"
#include "fold_vars.h"
#include "pair_mat.h"
/*@unused@*/
static char rcsid[] = "$Id: part_func.c,v 1.3 2003/12/17 13:23:03 rbund Exp $";

#define MAX(x,y) (((x)>(y)) ? (x) : (y))
#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#define PUBLIC
#define PRIVATE static
#define STACK_BULGE1  1   /* stacking energies for bulges of size 1 */
#define NEW_NINIO     1   /* new asymetry penalty */


PUBLIC  float pf_fold(char *sequence, char *structure);
PUBLIC  void  init_pf_fold(int length);
PUBLIC  void  free_pf_arrays(void);
PUBLIC  void  update_pf_params(int length);
PUBLIC  char  bppm_symbol(float *x);
PRIVATE void restricted_pf_fold(int n, FLT_OR_DBL *qr, int upper, int Delta,
				short *S, short *S1);
PRIVATE void  sprintf_bppm(int length, char *structure);
PRIVATE void  scale_pf_params(unsigned int length);
PRIVATE void  get_arrays(unsigned int length);
PRIVATE double expLoopEnergy(int u1, int u2, int type, int type2,
			     short si1, short sj1, short sp1, short sq1);
PRIVATE void make_ptypes(const short *S, const char *structure);

PRIVATE FLT_OR_DBL expMLclosing, expMLintern[NBPAIRS+1], *expMLbase;
PRIVATE FLT_OR_DBL expOUTintern[NBPAIRS+1], *expOUTbase, expOneOUTbase;
PRIVATE FLT_OR_DBL expTermAU;
PRIVATE FLT_OR_DBL expdangle5[NBPAIRS+1][5], expdangle3[NBPAIRS+1][5];
PRIVATE FLT_OR_DBL lxc, exptetra[40], expTriloop[40];
PRIVATE FLT_OR_DBL expstack[NBPAIRS+1][NBPAIRS+1];
PRIVATE FLT_OR_DBL expmismatchI[NBPAIRS+1][5][5],
  expmismatchH[NBPAIRS+1][5][5], expmismatchM[NBPAIRS+1][5][5];
PRIVATE FLT_OR_DBL expint11[NBPAIRS+1][NBPAIRS+1][5][5];
PRIVATE FLT_OR_DBL expint21[NBPAIRS+1][NBPAIRS+1][5][5][5];
PRIVATE FLT_OR_DBL expint22[NBPAIRS+1][NBPAIRS+1][5][5][5][5];
PRIVATE FLT_OR_DBL *exphairpin;
PRIVATE FLT_OR_DBL expbulge[MAXLOOP+1];
PRIVATE FLT_OR_DBL expinternal[MAXLOOP+1];
PRIVATE FLT_OR_DBL expninio[5][MAXLOOP+1];
PRIVATE FLT_OR_DBL *q,/* total partition function for subsequences, 
	also reused by Ralf as the weight that a strand of length j
	has extension i  */
	 *qb, *qm, *qqm, *qqm1, *qq, *qq1;
PRIVATE FLT_OR_DBL *prml, *prm_l, *prm_l1, *q1k,/*q1k[k]=q[1,k]*/ *qln;/*qln[k]=q[k,n]*/
PRIVATE FLT_OR_DBL *scale;
PRIVATE char *ptype; /* precomputed array of pair types */ 
PRIVATE int *jindx;
PRIVATE int init_length; /* length in last call to init_pf_fold() */
#define ISOLATED  256.0

PUBLIC float pf_fold(char *sequence, char *structure)
{
  short *S, *S1;
  int n/*number of bases in seq*/, i,j,k,l, ij, kl, u,u1,d,ii,ll, type, type_2, tt, ov=0;
  FLT_OR_DBL temp, Q, Qmax=0, prm_MLb;
  FLT_OR_DBL prmt,prmt1;
  FLT_OR_DBL qbt1, *tmp;
  FLT_OR_DBL maxq;
  FLT_OR_DBL newlq;
  FLT_OR_DBL qboundary;
   
  float free_energy;
  int Delta;   /* number of bases which have same size as a helix */

  n = (int) strlen(sequence);
  if (n>init_length) init_pf_fold(n);  /* (re)allocate space */

  S = (short *) space(sizeof(short)*(n+1));
  S1= (short *) space(sizeof(short)*(n+1));
  S[0] = n;
  for (l=1; l<=n; l++) {
    S[l]  = (short) encode_char(toupper(sequence[l-1]));
    S1[l] = alias[S[l]];
  }
  make_ptypes(S, structure);

  /* find boundary for automatic rescaling */
  if (sizeof(FLT_OR_DBL)==sizeof(float)) {
    qboundary=log(1.0e15);
  } else {
    qboundary=log(1.0e150);
  }
  
  /*array initialization ; qb,qm,q
    qb,qm,q (i,j) are stored as ((n+1-i)*(n-i) div 2 + n+1-j */

  for (d=0; d<=TURN; d++) 
    for (i=1; i<=n-d; i++) {
      j=i+d;
      ij = iindx[i]-j;
      q[ij]=1.0*scale[d+1];
      qb[ij]=qm[ij]=0.0;
    }

  for (i=1; i<=n; i++) 
    qq[i]=qq1[i]=qqm[i]=qqm1[i]=prm_l[i]=prm_l1[i]=prml[i]=0;

  for (j=TURN+2;j<=n; j++) {
    maxq=0.0;
    for (i=j-TURN-1; i>=1; i--) {
      /* construction of partition function of segment i,j*/
      /*firstly that given i bound to j : qb(i,j) */
      u = j-i-1; ij = iindx[i]-j;
      type = ptype[ij];
      if (type!=0) {
	/*hairpin contribution*/
	if (((type==3)||(type==4))&&no_closingGU) qbt1 = 0;
	else {
	  qbt1 = exphairpin[u];
	  if ((tetra_loop)&&(u==4)) {
	    char tl[7]={0}, *ts;
	    strncpy(tl, sequence+i-1, 6);
	    if ((ts=strstr(Tetraloops, tl)))
	      qbt1 *= exptetra[(ts-Tetraloops)/7];
	  } 
	  if (u==3) {
	    char tl[6]={0,0,0,0,0,0}, *ts;
	    strncpy(tl, sequence+i-1, 5);
	    if ((ts=strstr(Triloops, tl))) 
	      qbt1 *= expTriloop[(ts-Triloops)/6];
	    if (type>2) 
	      qbt1 *= expTermAU;
	  }
	  else /* no mismatches for tri-loops */
	    qbt1 *= expmismatchH[type][S1[i+1]][S1[j-1]];
	  
	  qbt1 *= scale[u+2];
	}
	/* interior loops with interior pair k,l */
	for (k=i+1; k<=MIN(i+MAXLOOP+1,j-TURN-2); k++) {
	  u1 = k-i-1;
	  for (l=MAX(k+TURN+1,j-1-MAXLOOP+u1); l<=j-1; l++) {
	    type_2 = ptype[iindx[k]-l];
	    if (type_2) {
	      type_2 = rtype[type_2];
	      qbt1 += qb[iindx[k]-l] * 
		expLoopEnergy(u1, j-l-1, type, type_2,
			      S1[i+1], S1[j-1], S1[k-1], S1[l+1]);
	    }
	  }
	}
	/*multiple stem loop contribution*/
	ii = iindx[i+1]; /* ii-k=[i+1,k-1] */
	temp = 0.0;
	for (k=i+2; k<=j-1; k++) temp += qm[ii-(k-1)]*qqm1[k]; 
	tt = rtype[type];
	qbt1 += temp*expMLclosing*expMLintern[tt]*scale[2]*
	  expdangle3[tt][S1[i+1]]*expdangle5[tt][S1[j-1]];
	 
	qb[ij] = qbt1;
	if (qbt1>0.0 && j-i>100) {
	  newlq=log(qbt1)/(j-i+1);
	  if (newlq>maxq) maxq=newlq;
	}
      } /* end if (type!=0) */
      else qb[ij] = 0.0;
       
      /* construction of qqm matrix containing final stem
	 contributions to multiple loop partition function
	 from segment i,j */
      qqm[i] = qqm1[i]*expMLbase[1];
      if (type) {
	qbt1 = qb[ij]*expMLintern[type];
	if (i>1) qbt1 *= expdangle5[type][S1[i-1]];
	if (j<n) qbt1 *= expdangle3[type][S1[j+1]];
	else if (type>2) qbt1 *= expTermAU;
	qqm[i] += qbt1;
      }
	 
      /*construction of qm matrix containing multiple loop
	partition function contributions from segment i,j */
      temp = 0.0;
      ii = iindx[i];  /* ii-k=[i,k-1] */
      for (k=i+1; k<=j; k++) temp += (qm[ii-(k-1)]+expMLbase[k-i])*qqm[k];
      qm[ij] = (temp + qqm[i]);
      if (qm[ij]>0.0 && j-i>100) {
	newlq=log(qm[ij])/(j-i+1);
	if (newlq>maxq) maxq=newlq;
      }
      
      /*auxiliary matrix qq for cubic order q calculation below */
      qbt1 = qb[ij]*expOUTintern[type];
      if (type) {
	if (i>1) qbt1 *= expdangle5[type][S1[i-1]];
	if (j<n) qbt1 *= expdangle3[type][S1[j+1]];
      }
      qq[i] = qq1[i]*expOUTbase[1] + qbt1;
      
      /*construction of partition function for segment i,j */
	/**********************************************************************/
      temp = 1.0*expOUTbase[1+j-i] + qq[i]; /*put NC energy here ?!?!?*/
	/**********************************************************************/
      for (k=i; k<=j-1; k++) temp += q[ii-k]*qq[k+1];
      q[ij] = temp;
      if (temp>0.0 && j-i>100) {
	newlq=log(temp)/(j-i+1);
	if (newlq>maxq) maxq=newlq;
      }

#ifndef LARGE_PF
	 if (temp>Qmax) {
	    Qmax = temp;
	    if (Qmax>FLT_MAX/10.)
	       fprintf(stderr, "%d %d %g\n", i,j,temp);
	 }
	 if (temp>FLT_MAX) {
	   PRIVATE char msg[128];
	    sprintf(msg, "overflow in pf_fold while calculating q[%d,%d]\n"
		    "use larger pf_scale", i,j);
	    nrerror(msg);
	 }
#endif
      }
      tmp = qq1;  qq1 =qq;  qq =tmp;
      tmp = qqm1; qqm1=qqm; qqm=tmp;

      /* check if we have to rescale everything */
      if (do_autoscale) {
	if ((maxq*j<0.25*qboundary && j>110) || maxq*j>qboundary) {
	      
	  FLT_OR_DBL Ratio;     /* ratio of pf_scales          */
	  long double RPower;   /* power of ratio of pf_scales */

	  /* rescale all the full length arrays */
	  if (maxq*j>qboundary) {
	    Ratio=exp(0.25*maxq);
	  } else {
	    Ratio=exp(-(0.5*qboundary/j-maxq));
	  }
	  expOneOUTbase/=Ratio;
	  scale[1]/=Ratio;
	  expMLbase[1]/=Ratio;
	  for(i=1;i<=n;i++) {
	    expOUTbase[i]=expOUTbase[i-1]*expOneOUTbase;
	    scale[i]=scale[i-1]*scale[1];
	    if (i<n) expMLbase[i]=expMLbase[i-1]*expMLbase[1];
	  }

	  /* rescale the two dimensional arrays */
	  RPower=Ratio;
	  for(i=0;i<=j;i++) {
	    for(k=1;k+i<=j;k++) {
	      q[iindx[k]-(k+i)]/=RPower;
	      qm[iindx[k]-(k+i)]/=RPower;
	      qb[iindx[k]-(k+i)]/=RPower;
	    }
	    RPower*=Ratio;
	  }

	  /* rescale the one dimensional arrays */
	  RPower=Ratio*Ratio;
	  for(i=1;i<j;i++) {
	    qq[j-i]/=RPower;
	    qq1[j-i]/=RPower;
	    qqm[j-i]/=RPower;
	    qqm1[j-i]/=RPower;
	    RPower*=Ratio;
	  }

	  pf_scale*=Ratio;
	  
	} /* end of rescaling necessary ? */
      } /* end of automatic rescaling requested ? */
    } /* end of loop over last base (j) */
   if (backtrack_type=='C')      Q = qb[iindx[1]-n];
   else if (backtrack_type=='M') Q = qm[iindx[1]-n];
   else Q = q[iindx[1]-n];

   /* ensemble free energy in Kcal/mol */
   if (Q<=FLT_MIN) fprintf(stderr, "pf_scale too large\n");
   free_energy = (-log(Q)-n*log(pf_scale))*(temperature+K0)*GASCONST/1000.0;
   /* in case we abort because of floating point errors */ 
   if (n>1600) fprintf(stderr, "free energy = %8.2f\n", free_energy); 


/************************************************************************/
/*			Fixed Extension Calculations		 	*/
/************************************************************************/


   /* we copy the parts of the array q which we might need later on */
   for (k=1; k<=n; k++) {
     q1k[k] = q[iindx[1] - k];
     qln[k] = q[iindx[k] -n];
   }
   q1k[0] = 1.0;
   qln[n+1] = 1.0;

   /* partition function restricted to a certain number of open bases */
   
   Delta=(int)(((float)OUT_intern)/((float)OUT_BASE)+0.5);
   if (Delta<0) Delta=0; /* we cannot work with Delta<0 */
   /* helices should not be bigger than their constituents */
   if (Delta>TURN+2) Delta=TURN+2;

   if (do_fixedextension || (open_bases>=0 && open_bases<=n)) {

     /* we will reuse q for our own purposes now                      */
     /* q[iindx[i]-j] is the weight that a strand of length j has     */
     /*               extension i                                     */
     /* note, that indeed q[iindx[i]-j] is valid for 0<=i<=j !        */

     fx=qqm;  /* recycle vector */

     /* this routine does all the work filling up the array q */
     restricted_pf_fold(n, q, 0, Delta, S, S1);
     
     for(i=0;i<=n;i++) {
       fx[i]=(-log(q[iindx[i]-n])-i*log(expOneOUTbase)-n*log(pf_scale))*
	 (temperature+K0)*GASCONST/1000.0;
     }

     /* check if restricted partition function for all lengthes are
	requested */
     if (return_alllengths) {

       /* we may have to allocate memory to be able to return this */
       if (!fmn) {
	 fmn = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));
       }

       if (fmn) {
	 for(i=0;i<=n;i++) {
	   for(j=0;j<=i;j++) {
	     fmn[iindx[j]-i]=(-log(q[iindx[j]-i])-(i-j)*log(pf_scale))*
	       (temperature+K0)*GASCONST/1000.0;
	   }
	 }
       }

       /* we need memory for the backwards part as well */
       
       if (!fmnb) {
	 fmnb = (FLT_OR_DBL *) space(sizeof(FLT_OR_DBL) *
				    ((init_length+1)*(init_length+2)/2+1));
       }

       if (fmnb) {
	 restricted_pf_fold(n, fmnb, 1, Delta, S, S1);
	 for(i=0;i<=n;i++) {
	   for(j=0;j<=i;j++) {
	     fmnb[iindx[j]-i]=(-log(fmnb[iindx[j]-i])-j*log(expOneOUTbase)
			       -i*log(pf_scale))*
	       (temperature+K0)*GASCONST/1000.0;
	   }
	 }
       }
       
     } /* end of: do we want all restricted partition functions ? */
     
   } /* end if (do_fixedextension || open_bases>=0) */

   /* in case open_bases asks for it, we replace the total free energy
      by the restricted one */
   if (open_bases>=0 && open_bases<=n) {
     if (backtrack_type=='C') {
       if (open_bases>0)
	 free_energy=-log(0.0);  /* impossible -> infinite free energy */
       /* for open_bases=0 the original assignment was good */
     } else
       free_energy=fx[open_bases];
   }
  

/************************************************************************/
/*			 BACKTRACKING 					*/
/************************************************************************/
 
  /* backtracking to construct binding probabilities of pairs*/
   
   if (do_backtrack) {
      int probabilities_vanish;
     
      Qmax=0;

    for (k=1; k<=n; k++) {
      q1k[k] = q[iindx[1] - k];
      qln[k] = q[iindx[k] -n];
    }
    q1k[0] = 1.0;
    qln[n+1] = 1.0;
      
    pr = q;     /* recycling */

    probabilities_vanish=0;
      
    /* 1. exterior pair i,j and initialization of pr array */
    if (open_bases>=0 && open_bases<=n) {/*number of open bases restricted?*/

      FLT_OR_DBL *qo;   /* array for restricted partition functions j...n */

      qo=calloc(sizeof(FLT_OR_DBL),(n+1)*(n+2)/2+1);
      if (!finite(fx[n]) || !finite(fx[open_bases]) || !qo) {

	/* we cannot calculate the base pairing probabilities */
	if (qo) free((void *)qo);
	fprintf(stderr,
		"Base pairing probabilities not available for %d open bases.\n",
		open_bases);
	for(i=1;i<=n;i++)
	  for(j=i;j<=n;j++)
	    pr[iindx[i]-j]=0.0;
	probabilities_vanish=1;
	  
      } else { /* no problem with calculation of base pairing matrix */

	/* full partition function for this number of open bases */
	FLT_OR_DBL qn;

	/* as a first thing we have to calculate the restricted
	   partition function of the substrand j...n under the
	   constraint that it has 0<=i<=n-j+1 exterior bases.
	   We store this partition function in qo[iindx[i]-(n-j+1)]. */
	
	restricted_pf_fold(n, qo, 1, Delta, S, S1);

	/* we are now ready to initialize the array of the base
	   pairing probabilities. This array is identical with the
	   array used for the restricted partition functions of
	   the substrands 1...j. Note, however, that in order
	   to fill pr[iindx[i]-j] for a given j we only need the
	   restricted partition functions 1...(j-1) which have not been
	   overwritten yet. We just have to save the total partition
	   function at the given number of exterior bases */
	qn=q[iindx[open_bases]-n];
	for (j=n; j>=1; j--) {
	  for (i=j; i>=MAX(j-TURN,1); i--) pr[iindx[i]-j] = 0;
	  for (i=j-TURN-1; i>=1; i--) {
	    ij = iindx[i]-j;
	    type = ptype[ij];
	    if (type&&(qb[ij]>0.)) {
	      /* add all combinations of the left and right strand which
		 give the correct number of open bases */
	      temp=0.0;
	      for(k=MAX(0,open_bases-Delta-n+j);
		  k<=MIN(open_bases-Delta,i-1);k++) {
		if (i==1)
		  if (j==n)
		    temp+=scale[open_bases-Delta]/expOUTbase[open_bases-Delta];
		  else
		    temp+=qo[iindx[open_bases-Delta-k]-(n-j)]*
		      scale[k]/expOUTbase[k];
		else if (j==n)
		  temp+=q[iindx[k]-(i-1)]*
		    scale[open_bases-Delta-k]/expOUTbase[open_bases-Delta-k];
		else
		  temp+=q[iindx[k]-(i-1)]*
		    qo[iindx[open_bases-Delta-k]-(n-j)];
	      }
	      pr[ij] = temp/qn*expOUTintern[type]/expOUTbase[Delta];
	      if (i>1) pr[ij] *= expdangle5[type][S1[i-1]];
	      if (j<n) pr[ij] *= expdangle3[type][S1[j+1]];
	    } else
	      pr[ij] = 0;
	  }
	}
	
	/* get rid of intermediary matrix */
	free((void *)qo);
      }
      
    } else {   /* number of open bases not restricted */
	
      for (i=1; i<=n; i++) {
	for (j=i; j<=MIN(i+TURN,n); j++) pr[iindx[i]-j] = 0;
	for (j=i+TURN+1; j<=n; j++) {
	  ij = iindx[i]-j;
	  type = ptype[ij];
	  if (type&&(qb[ij]>0.)) {
	    pr[ij] = q1k[i-1]*qln[j+1]/q1k[n];
	    if (i>1) pr[ij] *= expdangle5[type][S1[i-1]]*expOUTintern[type];
	    if (j<n) pr[ij] *= expdangle3[type][S1[j+1]];
	    else if (type>2) pr[ij] *= expTermAU;
	  } else
	    pr[ij] = 0;
	}
      }

    } /* end of: number of open bases restricted ? */

    /* loop over end bases if probabilities do not vanish completely */
      
    for (l=n; l>TURN+1 && !probabilities_vanish; l--) {
      
      /* 2. bonding k,l as substem of 2:loop enclosed by i,j */
      for (k=1; k<l-TURN; k++) {
	kl = iindx[k]-l;
	type_2 = ptype[kl]; type_2 = rtype[type_2];
	if (qb[kl]==0) continue;
	
	for (i=MAX(1,k-MAXLOOP-1); i<=k-1; i++) 
	  for (j=l+1; j<=MIN(l+ MAXLOOP -k+i+2,n); j++) {
	    ij = iindx[i] - j;
	    type = ptype[ij];
	    if ((pr[ij]>0)) {
	      pr[kl] += pr[ij]*expLoopEnergy(k-i-1, j-l-1, type, type_2,
					     S1[i+1], S1[j-1], S1[k-1], S1[l+1]);
	    } 
	  }
      }
      /* 3. bonding k,l as substem of multi-loop enclosed by i,j */
      prm_MLb = 0.;
      if (l<n) for (k=2; k<l-TURN; k++) {
	i = k-1;
	prmt = prmt1 = 0.0;
	    
	ii = iindx[i];     /* ii-j=[i,j]     */
	ll = iindx[l+1];   /* ll-j=[l+1,j-1] */
	tt = ptype[ii-(l+1)]; tt=rtype[tt];
	prmt1 = pr[ii-(l+1)]*expMLclosing*expMLintern[tt]*
	  expdangle3[tt][S1[i+1]]*expdangle5[tt][S1[l]];
	for (j=l+2; j<=n; j++) {
	  tt = ptype[ii-j]; tt = rtype[tt];
	  prmt += pr[ii-j]*expdangle3[tt][S1[i+1]]*
	    expdangle5[tt][S1[j-1]] *qm[ll-(j-1)];
	}
	kl = iindx[k]-l;
	tt = ptype[kl];
	prmt *= expMLclosing*expMLintern[tt];
	prml[ i] = prmt;
	prm_l[i] = prm_l1[i]*expMLbase[1]+prmt1;

	prm_MLb = prm_MLb*expMLbase[1] + prml[i];
	/* same as:    prm_MLb = 0;
	   for (i=1; i<=k-1; i++) prm_MLb += prml[i]*expMLbase[k-i-1]; */

	prml[i] = prml[ i] + prm_l[i];
	    
	if (qb[kl] == 0.) continue; 
	    
	temp = prm_MLb;

	for (i=1;i<=k-2; i++) 
	  temp += prml[i]*qm[iindx[i+1] - (k-1)];

	temp *= expMLintern[tt]*scale[2];
	if (k>1) temp *= expdangle5[tt][S1[k-1]];
	if (l<n) temp *= expdangle3[tt][S1[l+1]];
	pr[kl] += temp;
#ifndef LARGE_PF
	if (pr[kl]>Qmax) {
	  Qmax = pr[kl];
	  if (Qmax>FLT_MAX/10.)
	    fprintf(stderr, "%d %d %g %g\n", i,j,pr[kl],qb[kl]);
	}
	if (pr[kl]>FLT_MAX) {
	  ov++;
	  pr[kl]=FLT_MAX;
	}
#endif
      } /* end for (k=..) */
      tmp = prm_l1; prm_l1=prm_l; prm_l=tmp;

    }  /* end for (l=..)   */
      
    for (i=1; i<=n; i++)
      for (j=i+TURN+1; j<=n; j++) {
	ij = iindx[i]-j;
	pr[ij] *= qb[ij];
      }
      
    if (structure!=NULL)
      sprintf_bppm(n, structure);
  }   /* end if (do_backtrack)*/
   
  free(S);
  free(S1);
  if (ov>0) fprintf(stderr, "%d overflows occurred while backtracking;\n"
		    "you might try a smaller pf_scale than %g\n",
		    ov, pf_scale);
  /* assign partial free energies */
  for(i=1;i<=n;i++) {
    q1k[i] = (-log(q1k[i])-i*log(pf_scale))*(temperature+K0)*GASCONST/1000.0;
  }
  fl=q1k;
  
  return free_energy; 
}
